"""
The :mod:`ndlib.models.dynamic` module contains dynamic network models, for use with dynetx graph objects.
"""

from .DynKerteszThresholdModel import DynKerteszThresholdModel
from .DynProfileModel import DynProfileModel
from .DynProfileThresholdModel import DynProfileThresholdModel
from .DynSIModel import DynSIModel
from .DynSIRModel import DynSIRModel
from .DynSISModel import DynSISModel

__all__ = [
    'DynSIModel',
    'DynSIRModel',
    'DynSISModel',
    'DynKerteszThresholdModel',
    'DynProfileModel',
    'DynProfileThresholdModel',
]


import abc
import past.builtins
import networkx as nx
from ndlib.models.DiffusionModel import DiffusionModel
import six
import warnings
import numpy as np
import future.utils

__author__ = "Giulio Rossetti"
__license__ = "BSD-2-Clause"
__email__ = "giulio.rossetti@gmail.com"


class ConfigurationException(Exception):
    """Configuration Exception"""


@six.add_metaclass(abc.ABCMeta)
class DynamicDiffusionModel(DiffusionModel):
    """
        Partial Abstract Class that defines Diffusion Models
    """
    # __metaclass__ = abc.ABCMeta

    def __init__(self, graph, seed=None):
        """
            Model Constructor

            :param graph: A networkx graph object
        """

        np.random.seed(seed)

        self.discrete_state = True

        self.params = {
            'nodes': {},
            'edges': {},
            'model': {},
            'status': {}
        }

        self.available_statuses = {
            "Susceptible": 0,
            "Infected": 1,
            "Recovered": 2,
            "Blocked": -1
        }

        self.name = ""

        self.parameters = {
            "model": {},
            "nodes": {},
            "edges": {}
        }

        self.actual_iteration = 0
        self.dyngraph = graph
        self.graph = self.dyngraph
        self.status = {n: 0 for n in self.dyngraph.nodes()}
        self.initial_status = {}

        snapshot_ids = self.dyngraph.temporal_snapshots_ids()
        self.min_snapshot_id = min(snapshot_ids)
        self.max_snapshot_id = max(snapshot_ids)
        self.stream_execution = False

    def execute_snapshots(self, bunch_size=None, node_status=True):
        self.stream_execution = False
        if bunch_size is None:
            bunch_size = self.max_snapshot_id+1
        if bunch_size > self.max_snapshot_id+1:
            raise ValueError("Number of iterations required greater than snapshot number.")
        system_status = []
        for t in past.builtins.xrange(self.min_snapshot_id, bunch_size):
            self.graph = self.dyngraph.time_slice(t_from=t)
            its = self.iteration(node_status)
            system_status.append(its)
        return system_status

    def execute_iterations(self, node_status=True):
        self.stream_execution = True
        system_status = []
        for e in self.dyngraph.stream_interactions():
            if e[2] == '+':
                self.graph = nx.Graph()
                self.graph.add_edge(e[0], e[1])
                its = self.iteration(node_status)
                system_status.append(its)
        return system_status

    def set_initial_status(self, configuration):
        """
        Set the initial model configuration

        :param configuration: a ```ndlib.models.ModelConfig.Configuration``` object
        """

        self.__validate_configuration(configuration)

        nodes_cfg = configuration.get_nodes_configuration()
        # Set additional node information

        for param, node_to_value in future.utils.iteritems(nodes_cfg):
            if len(node_to_value) < len(self.graph.nodes()):
                raise ConfigurationException({"message": "Not all nodes have a configuration specified"})

            self.params['nodes'][param] = node_to_value

        edges_cfg = configuration.get_edges_configuration()
        # Set additional edges information
        for param, edge_to_values in future.utils.iteritems(edges_cfg):
            if len(edge_to_values) == len(self.graph.edges()):
                self.params['edges'][param] = {}
                for e in edge_to_values:
                    self.params['edges'][param][e] = edge_to_values[e]

        # Set initial status
        model_status = configuration.get_model_configuration()

        for param, nodes in future.utils.iteritems(model_status):
            self.params['status'][param] = nodes
            for node in nodes:
                self.status[node] = self.available_statuses[param]

        # Set model additional information
        model_params = configuration.get_model_parameters()
        for param, val in future.utils.iteritems(model_params):
            self.params['model'][param] = val

        # Handle initial infection
        if 'Infected' not in self.params['status']:
            if 'percentage_infected' in self.params['model']:
                number_of_initial_infected = len(self.graph.nodes()) * float(self.params['model']['percentage_infected'])
                if number_of_initial_infected < 1:
                    warnings.warn('Graph with less than 100 nodes: a single node will be set as infected')
                    number_of_initial_infected = 1

                available_nodes = [n for n in self.status if self.status[n] == 0]
                sampled_nodes = np.random.choice(available_nodes, int(number_of_initial_infected), replace=False)
                for k in sampled_nodes:
                    self.status[k] = self.available_statuses['Infected']

        self.initial_status = self.status

    def __validate_configuration(self, configuration):
        """
        Validate the consistency of a Configuration object for the specific model

        :param configuration: a Configuration object instance
        """
        if "Infected" not in self.available_statuses:
            raise ConfigurationException("'Infected' status not defined.")

        # Checking mandatory parameters
        omp = set([k for k in self.parameters['model'].keys() if not self.parameters['model'][k]['optional']])
        onp = set([k for k in self.parameters['nodes'].keys() if not self.parameters['nodes'][k]['optional']])
        oep = set([k for k in self.parameters['edges'].keys() if not self.parameters['edges'][k]['optional']])

        mdp = set(configuration.get_model_parameters().keys())
        ndp = set(configuration.get_nodes_configuration().keys())
        edp = set(configuration.get_edges_configuration().keys())

        if len(omp) > 0:
            if len(omp & mdp) != len(omp):
                raise ConfigurationException({"message": "Missing mandatory model parameter(s)", "parameters": omp-mdp})

        if len(onp) > 0:
            if len(onp & ndp) != len(onp):
                raise ConfigurationException({"message": "Missing mandatory node parameter(s)", "parameters": onp-ndp})

        if len(oep) > 0:
            if len(oep & edp) != len(oep):
                raise ConfigurationException({"message": "Missing mandatory edge parameter(s)", "parameters": oep-edp})

        # Checking optional parameters
        omp = set([k for k in self.parameters['model'].keys() if self.parameters['model'][k]['optional']])
        onp = set([k for k in self.parameters['nodes'].keys() if self.parameters['nodes'][k]['optional']])
        oep = set([k for k in self.parameters['edges'].keys() if self.parameters['edges'][k]['optional']])

        if len(omp) > 0:
            for param in omp:
                if param not in mdp:
                    configuration.add_model_parameter(param, self.parameters['model'][param]['default'])

        if len(onp) > 0:
            for param in onp:
                if param not in ndp:
                    for nid in self.graph.nodes():
                        configuration.add_node_configuration(param, nid, self.parameters['nodes'][param]['default'])

        if len(oep) > 0:
            for param in oep:
                if param not in edp:
                    for eid in self.graph.edges():
                        configuration.add_edge_configuration(param, eid, self.parameters['edges'][param]['default'])

        # Checking initial simulation status
        sts = set(configuration.get_model_configuration().keys())
        if self.discrete_state and "Infected" not in sts and "percentage_infected" not in mdp:
            warnings.warn('Initial infection missing: a random sample of 5% of graph nodes will be set as infected')
            self.params['model']["percentage_infected"] = 0.05

    iteration_bunch = execute_snapshots



"""
The :mod:`ndlib.models.opinions` module contains common opinion dynamics network models from research literature.
"""

__author__ = 'rossetti'
__license__ = "GPL"
__email__ = "giulio.rossetti@gmail.com"

from .AlgorithmicBiasModel import AlgorithmicBiasModel
from .CognitiveOpDynModel import CognitiveOpDynModel
from .MajorityRuleModel import MajorityRuleModel
from .QVoterModel import QVoterModel
from .SznajdModel import SznajdModel
from .VoterModel import VoterModel

__all__ = [
    'AlgorithmicBiasModel',
    'CognitiveOpDynModel',
    'MajorityRuleModel',
    'QVoterModel',
    'SznajdModel',
    'VoterModel',
]



from ndlib.viz.mpl.DiffusionViz import DiffusionPlot
import numpy as np

__author__ = 'Giulio Rossetti'
__license__ = "BSD-2-Clause"
__email__ = "giulio.rossetti@gmail.com"


class DiffusionPrevalence(DiffusionPlot):

    def __init__(self, model, trends):
        """
        :param model: The model object
        :param trends: The computed simulation iterations
        """
        super(self.__class__, self).__init__(model, trends)
        self.ylabel = "#Delta Nodes"
        self.title = "Prevalence"
        self.normalized = False

    def iteration_series(self, percentile):

        series = {k: [] for k in self.srev.keys()}

        presences = {k: [] for k in self.srev.keys()}
        for t in self.trends:

            for st in t:
                for k in t[st]['status_delta']:
                    presences[k].append(np.array(t[st]['status_delta'][k]))

        for st in presences:
            tp = np.percentile(np.array(presences[st]), percentile, axis=0)
            bp = np.percentile(np.array(presences[st]), 100 - percentile, axis=0)
            av = np.average(np.array(presences[st]), axis=0)
            series[st] = (tp, av, bp)

        return series



# TODO
# - Parallel execution
# - Add sensitivity analysis options

from SALib.sample import saltelli
from SALib.analyze import sobol
from ndlib.models.compartments.enums.SAType import SAType
import numpy as np


class ContinuousModelRunner(object):
    def __init__(self, ContinuousModel, config, node_status=True):
        self.model = ContinuousModel
        self.config = config
        self.node_status = node_status

    def run(self, N, iterations_list,  initial_statuses, constants_list=None):
        results = []

        n_iterations = len(iterations_list)
        n_initial_statuses = len(initial_statuses)

        if constants_list:
            n_constants = len(constants_list)

        for i in range(N):
            print('\nRunning simulation ' + str(i + 1) + '/' + str(N) + '\n')
            if constants_list:
                self.model.constants = constants_list[i % n_constants]
            self.model.set_initial_status(initial_statuses[i % n_initial_statuses], self.config)
            output = self.model.iteration_bunch(iterations_list[i % n_iterations], node_status=self.node_status)
            results.append(output)

        return results

    def analyze_sensitivity(self, sa_type, initial_status, bounds, n, iterations, second_order=True):
        """
        Compute the sensitivity indices for the constants of the model using sobol
        Samples are generated using a Saltelli sampler

        :param sa_type: SAType indicating the metric for the sensitivity analysis, possible values
        :param initial_status: a dictionary with as key a state and as value, a number or function indicating the intial value for this state
        :param bounds: a dictionary with a constant name as key and a tuple (lower bound, upper bound) as value
        :param n: integer indicating the n for the function: N∗(2D+2) which is used to determine the amount of samples
        :param iterations: amount of iterations to run the model per sample
        :param second_order: bool indicating whether to include second order indices

        :return: a Python dict mapping state to a dictionary with the keys "S1", "S2", "ST", "S1_conf", "S2_conf", and "ST_conf"
        """
        if not self.model.constants:
            raise Exception('Please add constants when initializing the model to perform sensitivity analysis on')
        if not isinstance(sa_type, SAType):
            raise ValueError('Please use a SAType enum value for sa_type')

        problem = {
            'num_vars': len(bounds.keys()),
            'names': [var for var in bounds.keys()],
            'bounds': [
                [lower, upper] for _, (lower, upper) in bounds.items()
            ]
        }

        param_values = saltelli.sample(problem, n, calc_second_order=second_order)

        outputs = []

        # Set the constants and run the model
        for i in range(len(param_values)):
            print('Running simulation ' + str(i + 1) + '/' + str(len(param_values)))
            # Set the constants
            for j, name in enumerate(problem['names']):
                self.model.constants[name] = param_values[i,j]
            # Set intial values
            self.model.set_initial_status(initial_status, self.config)
            outputs.append(self.model.iteration_bunch(iterations, node_status=self.node_status, tqdm=False))

        # Parse the outputs for every simulation (TODO: Optimize)
        print('Parsing outputs...')
        states = list(self.model.available_statuses.keys())
        states.remove('Infected')
        val_dict = {
            var: np.array([]) for var in states
        }
        values = self.parse_outputs(sa_type, outputs, val_dict)

        print('Running sensitivity analysis...')
        # Perform the sobol analysis seperately for every status
        analysis = {
            var: sobol.analyze(problem, values[var]) for var in states
        }

        return analysis

    def parse_outputs(self, sa_type, outputs, val_dict):
        mapping = {
            SAType.MEAN: self.mean_outputs
        }
        return mapping[sa_type](outputs, val_dict)

    def mean_outputs(self, outputs, val_dict):
        for output in outputs:
            # Get the average value of all the nodes for every status at the last iteration of the simulation
            means = self.model.get_means(output)
            for status in val_dict.keys():
                val_dict[status] = np.append(val_dict[status], means[status][-1])
        return val_dict

    def variance_outputs(self, outputs, val_dict):
        for output in outputs:
            # Get the average value of all the nodes for every status at the last iteration of the simulation
            means = self.model.get_means(output)
            for status in val_dict.keys():
                val_dict[status] = np.append(val_dict[status], means[status][-1])
        return val_dict


"""
The :mod:`ndlib.models.epidemic` module contains common network models from epidemic research literature.
"""

from .GeneralisedThresholdModel import GeneralisedThresholdModel
from .IndependentCascadesModel import IndependentCascadesModel
from .KerteszThresholdModel import KerteszThresholdModel
from .ProfileModel import ProfileModel
from .ProfileThresholdModel import ProfileThresholdModel
from .SEIRModel import SEIRModel
from .SEISModel import SEISModel
from .SIModel import SIModel
from .SIRModel import SIRModel
from .SISModel import SISModel
from .SWIRModel import SWIRModel
from .ThresholdModel import ThresholdModel
from .ICEModel import ICEModel
from .ICPModel import ICPModel
from .ICEPModel import ICEPModel
from .GeneralThresholdModel import GeneralThresholdModel
from .UTLDRModel import UTLDRModel
from .SEIR_ct_Model import SEIRctModel
from .SEIS_ct_Model import SEISctModel

__all__ = [
    'GeneralisedThresholdModel',
    'IndependentCascadesModel',
    'KerteszThresholdModel',
    'ProfileModel',
    'ProfileThresholdModel',
    'SEIRModel',
    'SEISModel',
    'SIModel',
    'SIRModel',
    'SISModel',
    'SWIRModel',
    'ThresholdModel',
    'ICEModel',
    'ICPModel',
    'GeneralThresholdModel',
    'UTLDRModel',
    'ICEPModel',
    'SEIRctModel',
    'SEISctModel'
]



import abc
import six

__author__ = 'Giulio Rossetti'
__license__ = "BSD-2-Clause"
__email__ = "giulio.rossetti@gmail.com"


class ConfigurationException(Exception):
    """Configuration Exception"""


@six.add_metaclass(abc.ABCMeta)
class Compartiment(object):
    """
    """

    def __init__(self, *args, **kwargs):
        self.composed = None
        if 'composed' in args[0]:
            if isinstance(args[0]['composed'], Compartiment):
                self.composed = args[0]['composed']

    def execute(self, *args, **kwargs):
        pass

    def compose(self, *args, **kwargs):
        if self.composed is not None:
            return self.composed.execute(*args, **kwargs)
        else:
            return True



from ndlib.models.compartments.Compartment import Compartiment, ConfigurationException

__author__ = 'Giulio Rossetti'
__license__ = "BSD-2-Clause"
__email__ = "giulio.rossetti@gmail.com"


class ConditionalComposition(Compartiment):

    def __init__(self, condition, first_branch, second_branch, **kwargs):
        super(self.__class__, self).__init__(kwargs)
        if not isinstance(condition, Compartiment) or not isinstance(first_branch, Compartiment) \
                or not isinstance(second_branch, Compartiment):
            raise ConfigurationException("Condition and branches must be of type Compartment.")

        self.condition = condition
        self.first_branch = first_branch
        self.second_branch = second_branch

    def execute(self, *args, **kwargs):
        test = self.condition.execute(*args, **kwargs)
        if test:
            return self.first_branch.execute(*args, **kwargs)
        else:
            return self.second_branch.execute(*args, **kwargs)



from ..DiffusionModel import DiffusionModel
import numpy as np
import future.utils

__author__ = "Alina Sirbu"
__email__ = "alina.sirbu@unipi.it"


class CognitiveOpDynModel(DiffusionModel):
    """
    Model Parameters to be specified via ModelConfig

    :param I: external information value in [0,1]
    :param T_range_min: the minimum of the range of initial values for  T. Range [0,1].
    :param T_range_max: the maximum of the range of initial values for  T. Range [0,1].
    :param B_range_min: the minimum of the range of initial values for  B. Range [0,1]
    :param B_range_max: the maximum of the range of initial values for  B. Range [0,1].
    :param R_fraction_negative: fraction of individuals having the node parameter R=-1.
    :param R_fraction_positive: fraction of individuals having the node parameter R=1
    :param R_fraction_neutral: fraction of individuals having the node parameter R=0

    The following relation should hold: R_fraction_negative+R_fraction_neutral+R_fraction_positive=1.
    To achieve this, the fractions selected will be normalised to sum 1.
    Node states are continuous values in [0,1].

    The initial state is generated randomly uniformly from the domain defined by model parameters.
    """

    def __init__(self, graph, seed=None):
        """
             Model Constructor

             :param graph: A networkx graph object
         """
        super(self.__class__, self).__init__(graph, seed)

        self.discrete_state = False

        self.available_statuses = {
            "Infected": 0
        }

        self.parameters = {
            "model": {
                "I": {
                    "descr": "External information",
                    "range": [0, 1],
                    "optional": False
                },
                "T_range_min": {
                    "descr": "Minimum of the range of initial values for T",
                    "range": [0, 1],
                    "optional": False
                },
                "T_range_max": {
                    "descr": "Maximum of the range of initial values for T",
                    "range": [0, 1],
                    "optional": False
                },
                "B_range_min": {
                    "descr": "Minimum of the range of initial values for B",
                    "range": [0, 1],
                    "optional": False
                },
                "B_range_max": {
                    "descr": "Maximum of the range of initial values for B",
                    "range": [0, 1],
                    "optional": False
                },
                "R_fraction_negative": {
                    "descr": "Fraction of nodes having R=-1",
                    "range": [0, 1],
                    "optional": False
                },
                "R_fraction_neutral": {
                    "descr": "Fraction of nodes having R=0",
                    "range": [0, 1],
                    "optional": False
                },
                "R_fraction_positive": {
                    "descr": "Fraction of nodes having R=1",
                    "range": [0, 1],
                    "optional": False
                }
            },
            "nodes": {},
            "edges": {}
        }

        self.name = "Cognitive Opinion Dynamics"

    def set_initial_status(self, configuration=None):
        """
        Override behaviour of methods in class DiffusionModel.
        Overwrites initial status using random real values.
        Generates random node profiles.
        """
        super(CognitiveOpDynModel, self).set_initial_status(configuration)

        # set node status
        for node in self.status:
            self.status[node] = np.random.random_sample()
        self.initial_status = self.status.copy()

        # set new node parameters
        self.params['nodes']['cognitive'] = {}

        # first correct the input model parameters and retreive T_range, B_range and R_distribution
        T_range = (self.params['model']['T_range_min'], self.params['model']['T_range_max'])
        if self.params['model']['T_range_min'] > self.params['model']['T_range_max']:
            T_range = (self.params['model']['T_range_max'], self.params['model']['T_range_min'])

        B_range = (self.params['model']['B_range_min'], self.params['model']['B_range_max'])
        if self.params['model']['B_range_min'] > self.params['model']['B_range_max']:
            B_range = (self.params['model']['B_range_max'], self.params['model']['B_range_min'])
        s = float(self.params['model']['R_fraction_negative'] + self.params['model']['R_fraction_neutral'] +
                  self.params['model']['R_fraction_positive'])
        R_distribution = (self.params['model']['R_fraction_negative']/s, self.params['model']['R_fraction_neutral']/s,
                          self.params['model']['R_fraction_positive']/s)

        # then sample parameters from the ranges and distribution
        for node in self.graph.nodes:
            R_prob = np.random.random_sample()
            if R_prob < R_distribution[0]:
                R = -1
            elif R_prob < (R_distribution[0] + R_distribution[1]):
                R = 0
            else:
                R = 1
            # R, B and T parameters in a tuple
            self.params['nodes']['cognitive'][node] = (R,
                                                       B_range[0] + (B_range[1] - B_range[0])*np.random.random_sample(),
                                                       T_range[0] + (T_range[1] - T_range[0])*np.random.random_sample())

    def clean_initial_status(self, valid_status=None):
        for n, s in future.utils.iteritems(self.status):
            if s > 1 or s < 0:
                self.status[n] = 0

    def iteration(self, node_status=True):
        """
        Execute a single model iteration

        :return: Iteration_id, Incremental node status (dictionary node->status)
        """
        # One iteration changes the opinion of all agents using the following procedure:
        # - first all agents communicate with institutional information I using a deffuant like rule
        # - then random pairs of agents are selected to interact  (N pairs)
        # - interaction depends on state of agents but also internal cognitive structure

        self.clean_initial_status(None)

        actual_status = {node: nstatus for node, nstatus in future.utils.iteritems(self.status)}

        if self.actual_iteration == 0:
            self.actual_iteration += 1
            delta, node_count, status_delta = self.status_delta(self.status)
            if node_status:
                return {"iteration": 0, "status": self.status.copy(),
                        "node_count": node_count.copy(), "status_delta": status_delta.copy()}
            else:
                return {"iteration": 0, "status": {},
                        "node_count": node_count.copy(), "status_delta": status_delta.copy()}

        # first interact with I
        I = self.params['model']['I']
        for node in self.graph.nodes:
            T = self.params['nodes']['cognitive'][node][2]
            R = self.params['nodes']['cognitive'][node][0]
            actual_status[node] = actual_status[node] + T * (I - actual_status[node])
            if R == 1:
                actual_status[node] = 0.5 * (1 + actual_status[node])
            if R == -1:
                actual_status[node] *= 0.5

        # then interact with peers
        for i in range(0, self.graph.number_of_nodes()):
            # select a random node
            n1 = list(self.graph.nodes)[np.random.randint(0, self.graph.number_of_nodes())]

            # select all of the nodes neighbours (no digraph possible)
            neighbours = list(self.graph.neighbors(n1))
            if len(neighbours) == 0:
                continue

            # select second node - a random neighbour
            n2 = neighbours[np.random.randint(0, len(neighbours))]

            # update status of n1 and n2
            p1 = pow(actual_status[n1], 1.0 / self.params['nodes']['cognitive'][n1][1])
            p2 = pow(actual_status[n2], 1.0 / self.params['nodes']['cognitive'][n2][1])

            oldn1 = self.status[n1]
            if np.random.random_sample() < p2:  # if node 2 talks, node 1 gets changed
                T1 = self.params['nodes']['cognitive'][n1][2]
                R1 = self.params['nodes']['cognitive'][n1][0]
                actual_status[n1] += (1 - T1) * (actual_status[n2] - actual_status[n1])
                if R1 == 1:
                    actual_status[n1] = 0.5 * (1 + actual_status[n1])
                if R1 == -1:
                    actual_status[n1] *= 0.5
            if np.random.random_sample() < p1:  # if node 1 talks, node 2 gets changed
                T2 = self.params['nodes']['cognitive'][n2][2]
                R2 = self.params['nodes']['cognitive'][n2][0]
                actual_status[n2] += (1 - T2) * (oldn1 - actual_status[n2])
                if R2 == 1:
                    actual_status[n2] = 0.5 * (1 + actual_status[n2])
                if R2 == -1:
                    actual_status[n2] *= 0.5

        delta, node_count, status_delta = self.status_delta(actual_status)
        self.status = actual_status
        self.actual_iteration += 1

        if node_status:
            return {"iteration": self.actual_iteration - 1, "status": delta.copy(),
                    "node_count": node_count.copy(), "status_delta": status_delta.copy()}
        else:
            return {"iteration": self.actual_iteration - 1, "status": {},
                    "node_count": node_count.copy(), "status_delta": status_delta.copy()}




from bokeh.layouts import gridplot

__author__ = 'Giulio Rossetti'
__license__ = "BSD-2-Clause"
__email__ = "giulio.rossetti@gmail.com"


class MultiPlot(object):

    def __init__(self):
        self.plots = []

    def add_plot(self, plot):
        """
        :param plot: The bokeh plot to add to the grid
        """
        self.plots.append(plot)

    def plot(self, ncols=2):
        """
        :param ncols: Number of grid columns
        :return: a bokeh figure image
        """
        grid = gridplot(self.plots, ncols=ncols)
        return grid



from ..DiffusionModel import DiffusionModel
import numpy as np
import future.utils
from scipy import stats

__author__ = "Giulio Rossetti"
__license__ = "BSD-2-Clause"
__email__ = "giulio.rossetti@gmail.com"


class ProfileThresholdModel(DiffusionModel):
    """
        Node Parameters to be specified via ModelConfig

        :param profile: The node profile. As default a value of 0.1 is assumed for all nodes.
        :param threshold: The node threshold. As default a value of 0.1 is assumed for all nodes.
    """

    def __init__(self, graph, seed=None):
        """
             Model Constructor

             :param graph: A networkx graph object
         """
        super(self.__class__, self).__init__(graph, seed)
        self.available_statuses = {
            "Susceptible": 0,
            "Infected": 1,
            "Blocked": -1
        }

        self.parameters = {
            "model": {
                "blocked": {
                    "descr": "Presence of blocked nodes",
                    "range": [0, 1],
                    "optional": True,
                    "default": 0
                },
                "adopter_rate": {
                    "descr": "Exogenous adoption rate",
                    "range": [0, 1],
                    "optional": True,
                    "default": 0
                }
            },
            "nodes": {
                "threshold": {
                    "descr": "Node threshold",
                    "range": [0, 1],
                    "optional": True,
                    "default": 0.1
                },
                "profile": {
                    "descr": "Node profile",
                    "range": [0, 1],
                    "optional": True,
                    "default": 0.1
                }
            },
            "edges": {},
        }

        self.name = "Profile-Threshold"

    def iteration(self, node_status=True):
        """
        Execute a single model iteration

        :return: Iteration_id, Incremental node status (dictionary node->status)
        """
        self.clean_initial_status(self.available_statuses.values())
        actual_status = {node: nstatus for node, nstatus in future.utils.iteritems(self.status)}

        if self.actual_iteration == 0:
            self.actual_iteration += 1
            delta, node_count, status_delta = self.status_delta(actual_status)
            if node_status:
                return {"iteration": 0, "status": actual_status.copy(),
                        "node_count": node_count.copy(), "status_delta": status_delta.copy()}
            else:
                return {"iteration": 0, "status": {},
                        "node_count": node_count.copy(), "status_delta": status_delta.copy()}

        for u in self.graph.nodes:
            if actual_status[u] != 0:
                continue

            if self.params['model']['adopter_rate'] > 0:
                xk = (0, 1)
                pk = (1 - self.params['model']['adopter_rate'], self.params['model']['adopter_rate'])
                probability = stats.rv_discrete(name='probability', values=(xk, pk))
                number_probability = probability.rvs()

                if number_probability == 1:
                    actual_status[u] = 1
                    continue

            neighbors = list(self.graph.neighbors(u))
            if self.graph.directed:
                neighbors = list(self.graph.predecessors(u))

            infected = 0
            for v in neighbors:
                infected += self.status[v]

            if infected > 0 and actual_status[u] == 0:

                infected_ratio = float(infected) / len(neighbors)
                if infected_ratio >= self.params['nodes']['threshold'][u]:
                    eventp = np.random.random_sample()
                    if eventp >= self.params['nodes']['profile'][u]:
                        actual_status[u] = 1
                    else:
                        if self.params['model']['blocked'] != 0:
                            blip = np.random.random_sample()
                            if blip > self.params['model']['blocked']:
                                actual_status[u] = -1

        delta, node_count, status_delta = self.status_delta(actual_status)
        self.status = actual_status
        self.actual_iteration += 1

        if node_status:
            return {"iteration": self.actual_iteration - 1, "status": delta.copy(),
                    "node_count": node_count.copy(), "status_delta": status_delta.copy()}
        else:
            return {"iteration": self.actual_iteration - 1, "status": {},
                    "node_count": node_count.copy(), "status_delta": status_delta.copy()}



from ndlib.models.DiffusionModel import DiffusionModel
import numpy as np
import future.utils

__author__ = 'Letizia Milli'
__license__ = "BSD-2-Clause"
__email__ = "milli@di.unipi.it"


class ICEPModel(DiffusionModel):
    """
        Edge Parameters to be specified via ModelConfig

        :param permeability: The degree of permeability of a community toward outgoing diffusion processes
    """

    def __init__(self, graph):
        """
             Model Constructor

             :param graph: A networkx graph object
         """
        super(self.__class__, self).__init__(graph)
        self.available_statuses = {
            "Susceptible": 0,
            "Infected": 1,
            "Removed": 2
        }

        self.parameters = {
            "model": {
                "permeability":{
                    "descr": "Community permeability",
                    "range": [0,1],
                    "optional": False,
                    "default": 0.5
                }
            },
            "nodes": {},
            "edges": {
                "threshold": {
                    "descr": "Edge threshold",
                    "range": [0, 1],
                    "optional": True,
                    "default": 0.1}
            },
        }

        self.name = "Community Permeability"

    def iteration(self, node_status=True):
        """
        Execute a single model iteration

        :return: Iteration_id, Incremental node status (dictionary node->status)
        """
        self.clean_initial_status(self.available_statuses.values())
        actual_status = {node: nstatus for node, nstatus in future.utils.iteritems(self.status)}

        if self.actual_iteration == 0:
            self.actual_iteration += 1
            delta, node_count, status_delta = self.status_delta(actual_status)
            if node_status:
                return {"iteration": 0, "status": actual_status.copy(),
                        "node_count": node_count.copy(), "status_delta": status_delta.copy()}
            else:
                return {"iteration": 0, "status": {},
                        "node_count": node_count.copy(), "status_delta": status_delta.copy()}

        edge_embeddedness = {}
        for u in self.graph.nodes:
            if self.status[u] != 1:
                continue

            edge_embeddedness[u] = {}

            neighbors = list(self.graph.neighbors(u))  # neighbors and successors (in DiGraph) produce the same result


              # Standard threshold
            if len(neighbors) > 0:
                threshold = 1.0/len(neighbors)

                for v in neighbors:
                    if actual_status[v] == 0:
                        key = (u, v)

                        same_community_neighbors = 0
                        neighbors_v = list(self.graph.neighbors(v))
                        for neighbor in neighbors_v:
                            if neighbor in neighbors and self.params['nodes']['com'][neighbor] == self.params['nodes']['com'][u]:
                                same_community_neighbors += 1
                        edge_embeddedness[u][v] = float(same_community_neighbors) / float(
                            len(neighbors) + len(neighbors_v) - same_community_neighbors)

                        if self.params['nodes']['com'][u] == self.params['nodes']['com'][v]:  # within same community
                            threshold = edge_embeddedness[u][v]  # Embedness

                        else:  # across communities
                            p = self.params['model']['permeability']
                            if 'threshold' in self.params['edges']:
                                if key in self.params['edges']['threshold']:
                                    threshold = self.params['edges']['threshold'][key] * p
                                elif (v, u) in self.params['edges']['threshold'] and not self.graph.directed:
                                    threshold = self.params['edges']['threshold'][(v, u)] * p

                        flip = np.random.random_sample()
                        if flip <= threshold:
                            actual_status[v] = 1

            actual_status[u] = 2

        delta, node_count, status_delta = self.status_delta(actual_status)
        self.status = actual_status
        self.actual_iteration += 1

        if node_status:
            return {"iteration": self.actual_iteration - 1, "status": delta.copy(),
                    "node_count": node_count.copy(), "status_delta": status_delta.copy()}
        else:
            return {"iteration": self.actual_iteration - 1, "status": {},
                    "node_count": node_count.copy(), "status_delta": status_delta.copy()}



import abc
from bokeh.plotting import figure
from bokeh.models import Title
from bokeh.palettes import Category20_9 as cols
import future.utils
import six

__author__ = 'Giulio Rossetti'
__license__ = "BSD-2-Clause"
__email__ = "giulio.rossetti@gmail.com"


@six.add_metaclass(abc.ABCMeta)
class DiffusionPlot(object):
    # __metaclass__ = abc.ABCMeta

    def __init__(self, model, trends):
        self.model = model
        self.trends = trends
        statuses = model.available_statuses
        self.srev = {v: k for k, v in future.utils.iteritems(statuses)}
        self.ylabel = ""
        self.title = ""
        self.nnodes = model.graph.number_of_nodes()
        self.normalized = True

    @abc.abstractmethod
    def iteration_series(self, percentile=100):
        """
        Prepare the data to be visualized

        :param percentile: The percentile for the trend variance area
        :return: a dictionary where iteration ids are keys and the associated values are the computed measures
        """
        pass

    def plot(self, percentile=100, width=500, height=500):
        """
        Generates the plot

        :param percentile: The percentile for the trend variance area
        :param width: Image width. Default 500px.
        :param height: Image height. Default 500px.
        :return: a bokeh figure image
        """
        pres = self.iteration_series(percentile)
        infos = self.model.get_info()
        descr = ""
        for k, v in future.utils.iteritems(infos):
            descr += "%s: %s, " % (k, v)
        descr = descr[:-2].replace("_", " ")

        p = figure(width=width, height=height)
        i = 0
        for k, l in future.utils.iteritems(pres):

            mx = len(l[0])
            if self.normalized:
                p.line(range(0, mx), l[1] / self.nnodes, line_width=2, legend=self.srev[k], alpha=0.5, color=cols[i])
            else:
                p.line(range(0, mx), l[1], line_width=2, legend=self.srev[k], alpha=0.5, color=cols[i])

            i += 1

        p.xaxis.axis_label = 'Iterations'
        p.title.text = "%s - %s" % (self.model.get_name(), self.title)
        p.yaxis.axis_label = self.ylabel
        p.ygrid[0].grid_line_alpha = 0.5
        p.xgrid[0].grid_line_alpha = 0.5
        p.add_layout(Title(text=descr, align="center"), "below")
        p.legend.orientation = "horizontal"

        return p


from ..DiffusionModel import DiffusionModel
import future.utils
import random
import queue
from collections import defaultdict


__author__ = "Letizia Milli"
__license__ = "BSD-2-Clause"
__email__ = "letizia.milli@isti.cnr.it"


class GeneralisedThresholdModel(DiffusionModel):
    """
        Node Parameters to be specified via ModelConfig
       :param threshold: The node threshold. If not specified otherwise a value of 0.1 is assumed for all nodes.
    """

    def __init__(self, graph, seed=None):
        """
             Model Constructor
             :param graph: A networkx graph object
         """
        super(self.__class__, self).__init__(graph, seed)
        self.available_statuses = {
            "Susceptible": 0,
            "Infected": 1
        }

        self.parameters = {
            "model": {
                "tau": {
                    "descr": "Adoption threshold rate",
                    "range": [0, float("inf")],
                    "optional": False},
                "mu": {
                    "descr": "Exogenous timescale",
                    "range": [0, float("inf")],
                    "optional": False},
            },
            "nodes": {
                "threshold": {
                    "descr": "Node threshold",
                    "range": [0, 1],
                    "optional": True,
                    "default": 0.1
                }
            },
            "edges": {},
        }

        self.queue = queue.PriorityQueue()
        self.inqueue = defaultdict(int)

        self.name = "GeneralisedThresholdModel"

    def iteration(self, node_status=True):
        """
        Execute a single model iteration
        :return: Iteration_id, Incremental node status (dictionary node->status)
        """
        self.clean_initial_status(self.available_statuses.values())

        actual_status = {node: nstatus for node, nstatus in future.utils.iteritems(self.status)}

        if self.actual_iteration == 0:
            self.actual_iteration += 1
            self.params['model']['queue'] = dict()
            return 0, actual_status

        gamma = float(self.params['model']['mu']) * float(self.actual_iteration) / float(self.params['model']['tau'])
        list_node = self.graph.nodes
        start = min(list_node)
        stop = max(list_node)
        number_node_susceptible = len(self.graph.nodes) - sum(self.status.values())

        while gamma >= 1 and number_node_susceptible >= 1:
            random_index = random.randrange(start, stop+1, 1)
            if random_index in list_node and actual_status[random_index] == 0:
                actual_status[random_index] = 1
                gamma -= 1
                number_node_susceptible -= 1

        for u in self.graph.nodes:
            if actual_status[u] == 1:
                continue

            neighbors = list(self.graph.neighbors(u))
            if self.graph.directed:
                neighbors = list(self.graph.predecessors(u))

            infected = 0
            for v in neighbors:
                infected += self.status[v]

            if len(neighbors) > 0:
                infected_ratio = float(infected)/len(neighbors)
                if infected_ratio >= self.params['nodes']['threshold'][u]:
                    if u not in self.inqueue:
                        self.queue.put((self.actual_iteration, u))
                        self.inqueue[u] = None

        while not self.queue.empty():
            next = self.queue.queue[0]
            if self.actual_iteration - next[0] >= self.params['model']['tau']:
                self.queue.get()
            else:
                break

        delta = self.status_delta(actual_status)
        self.status = actual_status
        self.actual_iteration += 1

        return self.actual_iteration - 1, delta



from ndlib.models.DiffusionModel import DiffusionModel
import numpy as np
import future.utils

__author__ = "Alina Sirbu"
__email__ = "alina.sirbu@unipi.it"


class AlgorithmicBiasModel(DiffusionModel):
    """
    Model Parameters to be specified via ModelConfig

    :param epsilon: bounded confidence threshold from the Deffuant model, in [0,1]
    :param gamma: strength of the algorithmic bias, positive, real

    Node states are continuous values in [0,1].

    The initial state is generated randomly uniformly from the domain [0,1].
    """

    def __init__(self, graph, seed=None):
        """
             Model Constructor

             :param graph: A networkx graph object
         """
        super(self.__class__, self).__init__(graph, seed)

        self.discrete_state = False

        self.available_statuses = {
            "Infected": 0
        }
        
        self.parameters = {
            "model": {
                "epsilon": {
                    "descr": "Bounded confidence threshold",
                    "range": [0, 1],
                    "optional": False
                },
                "gamma": {
                    "descr": "Algorithmic bias",
                    "range": [0, 100],
                    "optional": False
                }
            },
            "nodes": {},
            "edges": {}
        }

        self.name = "Agorithmic Bias"

    def set_initial_status(self, configuration=None):
        """
        Override behaviour of methods in class DiffusionModel.
        Overwrites initial status using random real values.
        """
        super(AlgorithmicBiasModel, self).set_initial_status(configuration)

        # set node status
        for node in self.status:
            self.status[node] = np.random.random_sample()
        self.initial_status = self.status.copy()

    def clean_initial_status(self, valid_status=None):
        for n, s in future.utils.iteritems(self.status):
            if s > 1 or s < 0:
                self.status[n] = 0

    @staticmethod
    def prob(distance, gamma, min_dist):
        if distance < min_dist:
            distance = min_dist
        return np.power(distance, -gamma)

    def iteration(self, node_status=True):
        """
        Execute a single model iteration

        :return: Iteration_id, Incremental node status (dictionary node->status)
        """
        # One iteration changes the opinion of N agent pairs using the following procedure:
        # - first one agent is selected
        # - then a second agent is selected based on a probability that decreases with the distance to the first agent
        # - if the two agents have a distance smaller than epsilon, then they change their status to the average of
        # their previous statuses

        self.clean_initial_status(None)

        actual_status = {node: nstatus for node, nstatus in future.utils.iteritems(self.status)}

        if self.actual_iteration == 0:
            self.actual_iteration += 1
            delta, node_count, status_delta = self.status_delta(self.status)
            if node_status:
                return {"iteration": 0, "status": self.status.copy(),
                        "node_count": node_count.copy(), "status_delta": status_delta.copy()}
            else:
                return {"iteration": 0, "status": {},
                        "node_count": node_count.copy(), "status_delta": status_delta.copy()}

        # interact with peers
        for i in range(0, self.graph.number_of_nodes()):
            # select a random node
            n1 = list(self.graph.nodes)[np.random.randint(0, self.graph.number_of_nodes())]
            # select all of the node's neighbours (no digraph possible)
            neighbours = list(self.graph.neighbors(n1))
            if len(neighbours) == 0:
                continue
        
            # compute probabilities to select a second node among the neighbours
            selection_prob = np.array([self.prob(np.abs(actual_status[neighbours[i]]-actual_status[n1]),
                                               self.params['model']['gamma'],0.00001) for i in range(len(neighbours))])
            selection_prob = selection_prob/np.sum(selection_prob)
            cumulative_selection_probability = np.cumsum(selection_prob)

            # select second nodebased on selection probabilities above
            r = np.random.random_sample()
            n2 = 0
            while cumulative_selection_probability[n2] < r:
                n2 = n2+1

            n2 = neighbours[n2]
            # update status of n1 and n2
            diff = np.abs(actual_status[n1]-actual_status[n2])
            if diff < self.params['model']['epsilon']:
                avg = (actual_status[n1]+actual_status[n2])/2.0
                actual_status[n1] = avg
                actual_status[n2] = avg
            
        delta, node_count, status_delta = self.status_delta(actual_status)
        self.status = actual_status
        self.actual_iteration += 1

        if node_status:
            return {"iteration": self.actual_iteration - 1, "status": delta.copy(),
                    "node_count": node_count.copy(), "status_delta": status_delta.copy()}
        else:
            return {"iteration": self.actual_iteration - 1, "status": {},
                    "node_count": node_count.copy(), "status_delta": status_delta.copy()}




from ndlib.models.compartments.Compartment import Compartiment
import networkx as nx
import numpy as np

__author__ = 'Giulio Rossetti'
__license__ = "BSD-2-Clause"
__email__ = "giulio.rossetti@gmail.com"


class NodeCategoricalAttribute(Compartiment):

    def __init__(self, attribute, value, probability=1, **kwargs):
        super(self.__class__, self).__init__(kwargs)

        if not isinstance(value, str):
            raise ValueError("Categorical (string) value expected")

        self.attribute = attribute
        self.attribute_value = value
        self.probability = probability

    def execute(self, node, graph, status, status_map, *args, **kwargs):

        val = nx.get_node_attributes(graph, self.attribute)[node]
        p = np.random.random_sample()
        test = val == self.attribute_value and p <= self.probability

        if test:
            return self.compose(node, graph, status, status_map, kwargs)

        return False



import abc
from bokeh.palettes import Category20_9 as cols
import os
import matplotlib as mpl
if os.environ.get('DISPLAY', '') == '':
    print('no display found. Using non-interactive Agg backend')
    mpl.use('Agg')
import matplotlib.pyplot as plt
import future.utils
import past
import six

__author__ = 'Giulio Rossetti'
__license__ = "BSD-2-Clause"
__email__ = "giulio.rossetti@gmail.com"


class InitializationException(Exception):
    """Initialization Exception"""


@six.add_metaclass(abc.ABCMeta)
class ComparisonPlot(object):
    # __metaclass__ = abc.ABCMeta

    def __init__(self, models, trends, statuses=["Infected"]):
        self.models = models
        self.trends = trends
        if len(models) != len(trends):
            raise InitializationException({"message": "The number of models does not match the number of trends"})

        sts = [model.available_statuses for model in models]
        self.mnames = ["%s_%s" % (models[i].name, i) for i in past.builtins.xrange(0, len(models))]
        self.srev = {}
        i = 0

        available_classes = {}
        for model in models:
            srev = {v: k for k, v in future.utils.iteritems(sts[i])}
            self.nnodes = model.graph.number_of_nodes()
            for cl in srev.values():
                available_classes[cl] = None

            self.srev["%s_%s" % (model.name, i)] = srev
            i += 1

        if type(statuses) == list:
            cls = set(statuses) & set(available_classes.keys())
        else:
            cls = set([statuses]) & set(available_classes.keys())
        if len(cls) > 0:
            self.classes = cls
        else:
            raise InitializationException({"message": "Statuses specified not available for the model (or missing)"})

        self.ylabel = ""
        self.title = ""
        self.normalized = True


    @abc.abstractmethod
    def iteration_series(self, percentile):
        """
        Prepare the data to be visualized

        :param percentile: The percentile for the trend variance area
        :return: a dictionary where iteration ids are keys and the associated values are the computed measures.
        """
        pass

    def plot(self, filename=None, percentile=90):
        """
        Plot the comparison on file.

        :param filename: the output filename
        :param percentile: The percentile for the trend variance area. Default 90.

        """

        pres = self.iteration_series(percentile)

        mx = 0
        i, h = 0, 0
        for k, l in future.utils.iteritems(pres):
            j = 0
            for st in l:
                mx = len(l[st][0])
                if self.normalized:
                    plt.plot(range(0, mx), l[st][1]/self.nnodes, lw=2,
                             label="%s - %s" % (k.split("_")[0], st), alpha=0.9, color=cols[h+j])
                    plt.fill_between(range(0,  mx), l[st][0]/self.nnodes,
                                     l[st][2]/self.nnodes, alpha=0.2, color=cols[h+j])
                else:
                    plt.plot(range(0, mx), l[st][1], lw=2,
                             label="%s - %s" % (k.split("_")[0], st), alpha=0.9, color=cols[h + j])
                    plt.fill_between(range(0, mx), l[st][0],
                                     l[st][2], alpha=0.2, color=cols[h + j])
                j += 1
            i += 1
            h += 2

        plt.grid(axis="y")
        plt.xlabel("Iterations", fontsize=24)
        plt.ylabel(self.ylabel, fontsize=24)
        plt.legend(loc="best", fontsize=18)
        plt.xlim((0, mx))

        if self.normalized:
            plt.ylim((0, 1))

        plt.tight_layout()
        
        if filename is not None:
            plt.savefig(filename)
            plt.clf()
        else:
            plt.show()



from ndlib.models.compartments.Compartment import Compartiment
import numpy as np

__author__ = 'Giulio Rossetti'
__license__ = "BSD-2-Clause"
__email__ = "giulio.rossetti@gmail.com"


class NodeStochastic(Compartiment):

    def __init__(self, rate, triggering_status=None, **kwargs):
        super(self.__class__, self).__init__(kwargs)
        self.rate = rate
        self.trigger = triggering_status

    def execute(self, node, graph, status, status_map, *args, **kwargs):
        neighbors = graph.neighbors(node)
        try:
            directed = graph.directed
        except AttributeError:
            directed = graph.is_directed()

        if directed:
            neighbors = graph.predecessors(node)

        p = np.random.random_sample()
        if self.trigger is None:
            triggered = 1
        else:
            triggered = 1 if len([v for v in neighbors if status[v] == status_map[self.trigger]]) > 0 else 0

        test = p < self.rate * triggered
        if test:
            return self.compose(node, graph, status, status_map, kwargs)

        return False



from ndlib.models.compartments.Compartment import Compartiment, ConfigurationException

__author__ = 'Giulio Rossetti'
__license__ = "BSD-2-Clause"
__email__ = "giulio.rossetti@gmail.com"


class NodeThreshold(Compartiment):

    def __init__(self, threshold=None, triggering_status=None, **kwargs):
        super(self.__class__, self).__init__(kwargs)
        self.threshold = threshold
        self.trigger = triggering_status

    def execute(self, node, graph, status, status_map, params, *args, **kwargs):
        neighbors = list(graph.neighbors(node))
        test = False

        try:
            directed = graph.directed
        except AttributeError:
            directed = graph.is_directed()

        if directed:
            neighbors = list(graph.predecessors(node))

        if self.trigger is None:
            triggered = 0
        else:
            triggered = len([v for v in neighbors if status[v] == status_map[self.trigger]])

        if len(list(neighbors)) > 0:
            infected_ratio = float(triggered) / len(neighbors)
            if 'threshold' in params['nodes']:
                test = infected_ratio >= params['nodes']['threshold'][node]
            else:
                if self.threshold is not None:
                    test = infected_ratio >= self.threshold
                else:
                    raise ConfigurationException("Threshold not specified")
        if test:
            return self.compose(node, graph, status, status_map, params, kwargs)

        return False



from ndlib.models.compartments.Compartment import Compartiment, ConfigurationException
import numpy as np

__author__ = 'Giulio Rossetti'
__license__ = "BSD-2-Clause"
__email__ = "giulio.rossetti@gmail.com"


class EdgeStochastic(Compartiment):

    def __init__(self, threshold=None, triggering_status=None, **kwargs):
        super(self.__class__, self).__init__(kwargs)
        self.threshold = threshold
        self.trigger = triggering_status
        if self.trigger is None:
            raise ConfigurationException("Triggering status not specified.")

    def execute(self, node, graph, status, status_map, params, *args, **kwargs):
        neighbors = list(graph.neighbors(node))

        try:
            directed = graph.directed
        except AttributeError:
            directed = graph.is_directed()

        if directed:
            neighbors = list(graph.predecessors(node))

        threshold = float(1)/len(neighbors)
        triggered = [v for v in neighbors if status[v] == status_map[self.trigger]]

        for n in triggered:
            if 'threshold' in params['edges']:
                key = (node, n)
                try:
                    threshold = params['edges']['threshold'][key]
                except KeyError:
                    threshold = params['edges']['threshold'][(n, node)]
            else:
                if self.threshold is not None:
                    threshold = self.threshold
            flip = np.random.random_sample()
            test = flip <= threshold

            if test:
                return self.compose(node, graph, status, status_map, params, kwargs)

        return False




"""
The :mod:`ndlib.models.compartments` module contains...
"""

from .NodeStochastic import NodeStochastic
from .NodeThreshold import NodeThreshold
from .NodeCategoricalAttribute import NodeCategoricalAttribute
from .NodeNumericalAttribute import NodeNumericalAttribute
from .NodeNumericalVariable import NodeNumericalVariable
from .EdgeStochastic import EdgeStochastic
from .EdgeCategoricalAttribute import EdgeCategoricalAttribute
from .EdgeNumericalAttribute import EdgeNumericalAttribute
from .ConditionalComposition import ConditionalComposition
from .CountDown import CountDown

__all__ = [
    'NodeStochastic',
    'NodeThreshold',
    'NodeCategoricalAttribute',
    'NodeNumericalAttribute',
    'NodeNumericalVariable',
    'EdgeStochastic',
    'EdgeCategoricalAttribute',
    'EdgeNumericalAttribute',
    'ConditionalComposition',
    'CountDown',
]



from ndlib.models.compartments.Compartment import Compartiment
from ndlib.models.compartments.enums.NumericalType import NumericalType
import networkx as nx
import numpy as np
import operator

__author__ = 'Mathijs Maijer'
__license__ = "BSD-2-Clause"
__email__ = "m.f.maijer@gmail.com"


class NodeNumericalVariable(Compartiment):

    def __init__(self, var, var_type=None, value=None, value_type=None, op=None, probability=1, **kwargs):
        super(NodeNumericalVariable, self).__init__(kwargs)
        self.__available_operators = {"==": operator.__eq__, "<": operator.__lt__,
                                      ">": operator.__gt__, "<=": operator.__le__,
                                      ">=": operator.__ge__, "!=": operator.__ne__,
                                      "IN": (operator.__ge__, operator.__le__)}

        self.variable = var
        self.variable_type = var_type
        self.value = value
        self.value_type = value_type
        self.probability = probability
        self.operator = op

        self.validate()

    def validate(self):
        if not isinstance(self.variable, str):
            raise ValueError("The variable should be a string pointing to an attribute or status")
        if self.variable_type is None:
            raise ValueError("A type must be provided for the variable")
        if not isinstance(self.variable_type, NumericalType):
            raise ValueError("The provided variable type is not valid")
        if self.value_type and not isinstance(self.value_type, NumericalType):
            raise ValueError("The provided value type is not valid")
        if self.value is None:
            raise ValueError("A value must be provided")

        if self.operator is not None and self.operator in self.__available_operators:
            if self.operator == "IN":
                if not isinstance(self.value, list) or \
                    not (isinstance(self.value[0], int) or isinstance(self.value[0], float)) or \
                    not (isinstance(self.value[1], int) or isinstance(self.value[1], float)) \
                    or self.value[1] < self.value[0]:
                    raise ValueError("A range list is required to test IN condition")
            else:
                if self.value_type is None:
                    if not isinstance(self.value, int):
                        if not isinstance(self.value, float):
                            raise ValueError("When no value type is defined, the value should be numerical")
                else:
                    if not isinstance(self.value, str):
                        raise ValueError("The value should be a string pointing to an attribute or status when value type is set")
        else:
            raise ValueError("The operator provided '%s' is not valid" % operator)

    def execute(self, node, graph, status, status_map, attributes=None, *args, **kwargs):
        if self.variable_type == NumericalType.STATUS:
            val = status[node][self.variable]
        elif self.variable_type == NumericalType.ATTRIBUTE:
            if attributes:
                val = attributes[node][self.variable]
            else:
                val = nx.get_node_attributes(graph, self.variable)[node]

        testVal = self.value

        if self.value_type == NumericalType.STATUS:
            testVal = status[node][self.value]
        elif self.value_type == NumericalType.ATTRIBUTE:
            if attributes:
                testVal = attributes[node][self.value]
            else:
                testVal = nx.get_node_attributes(graph, self.value)[node]

        p = np.random.random_sample()

        if self.operator == "IN":
            condition = self.__available_operators[self.operator][0](val, testVal[0]) and \
                        self.__available_operators[self.operator][1](val, testVal[1])
        else:
            condition = self.__available_operators[self.operator](val, testVal)

        test = condition and p <= self.probability

        if test:
            return self.compose(node, graph, status, status_map, kwargs)

        return False



import abc
from bokeh.palettes import Category20_9 as cols
import os
import matplotlib as mpl
if os.environ.get('DISPLAY', '') == '':
    print('no display found. Using non-interactive Agg backend')
    mpl.use('Agg')
import matplotlib.pyplot as plt
import future.utils
import six

__author__ = 'Giulio Rossetti'
__license__ = "BSD-2-Clause"
__email__ = "giulio.rossetti@gmail.com"


@six.add_metaclass(abc.ABCMeta)
class DiffusionPlot(object):
   # __metaclass__ = abc.ABCMeta

    def __init__(self, model, trends):
        self.model = model
        self.trends = trends
        statuses = model.available_statuses
        self.srev = {v: k for k, v in future.utils.iteritems(statuses)}
        self.ylabel = ""
        self.title = ""
        self.nnodes = model.graph.number_of_nodes()
        self.normalized = True

    @abc.abstractmethod
    def iteration_series(self, percentile):
        """
        Prepare the data to be visualized

        :param percentile: The percentile for the trend variance area
        :return: a dictionary where iteration ids are keys and the associated values are the computed measures
        """
        pass

    def plot(self, filename=None, percentile=90, statuses=None):
        """
        Generates the plot

        :param filename: Output filename
        :param percentile: The percentile for the trend variance area
        :param statuses: List of statuses to plot. If not specified all statuses trends will be shown.
        """

        pres = self.iteration_series(percentile)
        # infos = self.model.get_info()
        # descr = ""

        plt.figure(figsize=(20, 10))

        # for k, v in future.utils.iteritems(infos):
        #     descr += "%s: %s, " % (k, v)
        # descr = descr[:-2].replace("_", " ")

        mx = 0
        i = 0
        for k, l in future.utils.iteritems(pres):

            if statuses is not None and self.srev[k] not in statuses:
                continue
            mx = len(l[0])
            if self.normalized:
                plt.plot(range(0, mx), l[1]/self.nnodes, lw=2, label=self.srev[k], alpha=0.5)  # , color=cols[i])
                plt.fill_between(range(0,  mx), l[0]/self.nnodes, l[2]/self.nnodes, alpha=0.2)
                    #,color=cols[i])
            else:
                plt.plot(range(0, mx), l[1], lw=2, label=self.srev[k], alpha=0.5)  # , color=cols[i])
                plt.fill_between(range(0, mx), l[0], l[2], alpha=0.2)  # ,color=cols[i])

            i += 1

        plt.grid(axis="y")
        # plt.title(descr)
        plt.xlabel("Iterations", fontsize=24)
        plt.ylabel(self.ylabel, fontsize=24)
        plt.legend(loc="best", fontsize=18)
        plt.xlim((0, mx))

        plt.tight_layout()
        if filename is not None:
            plt.savefig(filename)
            plt.clf()
        else:
            plt.show()



from ..DiffusionModel import DiffusionModel
import numpy as np

__author__ = "Alina Sirbu"
__email__ = "alina.sirbu@unipi.it"


class QVoterModel(DiffusionModel):
    """
    Node Parameters to be specified via ModelConfig

    :param q: the number of neighbors that affect the opinion of a node
    """

    def __init__(self, graph, seed=None):
        """
             Model Constructor

             :param graph: A networkx graph object
         """
        super(self.__class__, self).__init__(graph, seed)
        self.available_statuses = {
            "Susceptible": 0,
            "Infected": 1
        }

        self.parameters = {"model": {
                "q": {
                    "descr": "Number of neighbours that affect the opinion of an agent",
                    "range": [0, len(self.graph.nodes)],
                    "optional": False
                }
            },
            "nodes": {},
            "edges": {}
        }

        self.name = "QVoter"

    def iteration(self, node_status=True):
        """
        Execute a single model iteration

        :return: Iteration_id, Incremental node status (dictionary node->status)
        """
        # One iteration changes the opinion of one voter using the following procedure:
        # - select randomly one voter (listener)
        # - select randomly q of its neighbours (speakers)
        # - if the q neighbours agree, the listener takes their opinion

        self.clean_initial_status(self.available_statuses.values())

        if self.actual_iteration == 0:
            self.actual_iteration += 1
            delta, node_count, status_delta = self.status_delta(self.status)
            if node_status:
                return {"iteration": 0, "status": self.status.copy(),
                        "node_count": node_count.copy(), "status_delta": status_delta.copy()}
            else:
                return {"iteration": 0, "status": {},
                        "node_count": node_count.copy(), "status_delta": status_delta.copy()}

                # select a random listener
        listener = list(self.graph.nodes)[np.random.randint(0, self.graph.number_of_nodes())]

        # select all of the listener's neighbours
        neighbours = list(self.graph.neighbors(listener))
        if self.graph.directed:
            # consider only the predecessors
            # assumed if a->b then b can be influenced by a
            neighbours = list(self.graph.predecessors(listener))

        # select q random neighbours (with repetitions)
        influence_group_state = [self.status[neighbours[i]]
                                 for i in np.random.randint(0, len(neighbours), self.params['model']['q'])]

        delta = {}
        # if all neighbours agree (either on 0 or on 1)
        status_delta = {st: 0 for st in self.available_statuses.values()}

        if sum(influence_group_state) == 0 or sum(influence_group_state) == len(influence_group_state):
            # update status of listener to either on of the neighbours selected
            delta[listener] = influence_group_state[0]
            self.status[listener] = influence_group_state[0]
            status_delta[self.status[listener]] += 1
            for x in self.available_statuses.values():
                if x != self.status[listener]:
                    status_delta[x] -= 1
        # fix
        node_count = {st: len([n for n in self.status if self.status[n] == st])
                      for st in self.available_statuses.values()}

        self.actual_iteration += 1

        if node_status:
            return {"iteration": self.actual_iteration - 1, "status": delta.copy(),
                    "node_count": node_count.copy(), "status_delta": status_delta.copy()}
        else:
            return {"iteration": self.actual_iteration - 1, "status": {},
                    "node_count": node_count.copy(), "status_delta": status_delta.copy()}




from ..DiffusionModel import DiffusionModel
import numpy as np
import future

__author__ = ["Giulio Rossetti"]
__license__ = "BSD-2-Clause"


class UTLDRModel(DiffusionModel):

    def __init__(self, graph, seed=None):

        super(self.__class__, self).__init__(graph, seed)

        self.params['nodes']['vaccinated'] = {n: False for n in self.graph.nodes}
        self.params['nodes']['tested'] = {n: False for n in self.graph.nodes}
        self.params['nodes']['ICU'] = {n: False for n in self.graph.nodes}
        self.params['nodes']['filtered'] = {n: [] for n in self.graph.nodes}
        self.icu_b = self.graph.number_of_nodes()
        self.lockdown = False

        self.name = "UTLDR"

        self.available_statuses = {
            "Susceptible": 0,
            "Exposed": 2,
            "Infected": 1,
            "Recovered": 3,
            "Identified_Exposed": 4,
            "Hospitalized_mild": 5,
            "Hospitalized_severe_ICU": 6,
            "Hospitalized_severe": 7,
            "Lockdown_Susceptible": 8,
            "Lockdown_Exposed": 9,
            "Lockdown_Infected": 10,
            "Dead": 11,
            "Vaccinated": 12,
        }

        self.parameters = {
            "model": {
                "sigma": {
                    "descr": "Incubation rate (1/expected iterations)",
                    "range": [0, 1],
                    "optional": False
                },
                "beta": {
                    "descr": "Infection rate (1/expected iterations)",
                    "range": [0, 1],
                    "optional": False
                },
                "gamma": {
                    "descr": "Recovery rate - Mild, Asymptomatic, Paucisymptomatic (1/expected iterations)",
                    "range": [0, 1],
                    "optional": False
                },
                "gamma_t": {
                    "descr": "Recovery rate - Severe in ICU (1/expected iterations)",
                    "range": [0, 1],
                    "optional": True,
                    "default": 0.6
                },
                "gamma_f": {
                    "descr": "Recovery rate - Severe not in ICU (1/expected iterations)",
                    "range": [0, 1],
                    "optional": True,
                    "default": 0.95
                },
                "omega": {
                    "descr": "Death probability - Mild, Asymptomatic, Paucisymptomatic",
                    "range": [0, 1],
                    "optional": True,
                    "default": 0
                },
                "omega_t": {
                    "descr": "Death probability - Severe in ICU",
                    "range": [0, 1],
                    "optional": True,
                    "default": 0
                },
                "omega_f": {
                    "descr": "Death probability - Severe not in ICU",
                    "range": [0, 1],
                    "optional": True,
                    "default": 0
                },
                "phi_e": {
                    "descr": "Testing probability if Exposed",
                    "range": [0, 1],
                    "optional": True,
                    "default": 0
                },
                "phi_i": {
                    "descr": "Testing probability if Infected",
                    "range": [0, 1],
                    "optional": True,
                    "default": 0
                },
                "kappa_e": {
                    "descr": "Test False Negative probability if Exposed",
                    "range": [0, 1],
                    "optional": True,
                    "default": 0.7
                },
                "kappa_i": {
                    "descr": "Test False Negative probability if Infected",
                    "range": [0, 1],
                    "optional": True,
                    "default": 0.9
                },
                "epsilon_e": {
                    "descr": "Social restriction due to quarantine (percentage of pruned edges)",
                    "range": [0, 1],
                    "optional": True,
                    "default": 1
                },
                "epsilon_l": {
                    "descr": "Social restriction due to lockdown (maximum household size)",
                    "range": [0, 1],
                    "optional": True,
                    "default": 1
                },
                "lambda": {
                    "descr": "Lockdown effectiveness (percentage of compliant individuals)",
                    "range": [0, 1],
                    "optional": True,
                    "default": 1
                },
                "mu": {
                    "descr": "Lockdown duration (1/expected iterations)",
                    "range": [0, 1],
                    "optional": True,
                    "default": 1
                },
                "p": {
                    "descr": "Probability of long-range interactions",
                    "range": [0, 1],
                    "optional": True,
                    "default": 0
                },
                "p_l": {
                    "descr": "Probability of long-range interactions if in Lockdown",
                    "range": [0, 1],
                    "optional": True,
                    "default": 0
                },
                "lsize": {
                    "descr": "Percentage of long-range interactions w.r.t short-range ones",
                    "range": [0, 1],
                    "optional": True,
                    "default": 0.25
                },
                "icu_b": {
                    "descr": "Beds availability in ICU (absolute value)",
                    "range": [0, np.infty],
                    "optional": True,
                    "default": self.graph.number_of_nodes()
                },
                "iota": {
                    "descr": "Severe case probability (needing ICU treatments)",
                    "range": [0, 1],
                    "optional": True,
                    "default": 1
                },
                "z": {
                    "descr": "Probability of infection from corpses",
                    "range": [0, 1],
                    "optional": True,
                    "default": 0
                },
                "s": {
                    "descr": "Probability of absent immunization",
                    "range": [0, 1],
                    "optional": True,
                    "default": 0
                },
                "v": {
                    "descr": "Probability of vaccination (single chance per agent)",
                    "range": [0, 1],
                    "optional": True,
                    "default": 0
                },
                "f": {
                    "descr": "Probability of Vaccination nullification (inverse of temporal coverage)",
                    "range": [0, 1],
                    "optional": True,
                    "default": 0
                },
            },
            "nodes": {
                "activity": {
                    "descr": "Node interactions per iteration (sample size of existing social relationships, "
                             "with replacement)",
                    "range": [0, 1],
                    "optional": True,
                    "default": 1
                },
                "segment": {
                    "descr": "Node class (e.g., age, gender)",
                    "range": str,
                    "optional": True,
                    "default": None
                },
            },
            "edges": {},
        }

    def iteration(self, node_status=True):
        """

        :param node_status:
        :return:
        """

        self.clean_initial_status(self.available_statuses.values())

        actual_status = {node: nstatus for node, nstatus in future.utils.iteritems(self.status)}

        if self.actual_iteration == 0:
            self.icu_b = self.params['model']['icu_b']

            self.actual_iteration += 1
            delta, node_count, status_delta = self.status_delta(actual_status)
            if node_status:
                return {"iteration": 0, "status": actual_status.copy(),
                        "node_count": node_count.copy(), "status_delta": status_delta.copy()}
            else:
                return {"iteration": 0, "status": {},
                        "node_count": node_count.copy(), "status_delta": status_delta.copy()}

        for u in self.graph.nodes:

            u_status = self.status[u]

            if self.graph.directed:
                neighbors = self.graph.predecessors(u)
            else:
                neighbors = self.graph.neighbors(u)

            # checking social limitations due to quarantine and/or lockdown
            if self.params['nodes']['filtered'][u] is not None and len(self.params['nodes']['filtered'][u]) > 0:
                neighbors = list(set(neighbors) & set(self.params['nodes']['filtered'][u]))

            ####################### Undetected Compartment ###########################

            if u_status == self.available_statuses['Susceptible']:
                actual_status[u] = self.__Susceptible_to_Exposed(u, neighbors, lockdown=False)

            elif u_status == self.available_statuses['Exposed']:

                tested = np.random.random_sample()  # selection for testing
                if not self.params['nodes']['tested'][u] and tested < self.__get_threshold(u, 'phi_e'):
                    res = np.random.random_sample()  # probability of false negative result
                    if res > self.__get_threshold(u, 'kappa_e'):
                        self.__limit_social_contacts(u, neighbors, 'Tested')
                        actual_status[u] = self.available_statuses['Identified_Exposed']
                    self.params['nodes']['tested'][u] = True
                else:
                    at = np.random.random_sample()
                    if at < self.__get_threshold(u, 'sigma'):
                        actual_status[u] = self.available_statuses['Infected']

            elif u_status == self.available_statuses['Infected']:

                tested = np.random.random_sample()  # selection for testing
                if not self.params['nodes']['tested'][u] and tested < self.__get_threshold(u, 'phi_i'):
                    res = np.random.random_sample()  # probability of false negative result
                    if res > self.__get_threshold(u, 'kappa_i'):
                        self.__limit_social_contacts(u, neighbors, 'Tested')

                        icup = np.random.random_sample()  # probability of severe case needing ICU

                        if icup < self.__get_threshold(u, 'iota'):
                            if self.icu_b > 0 and not self.params['nodes']['ICU'][u]:
                                actual_status[u] = self.available_statuses['Hospitalized_severe_ICU']
                                self.icu_b -= 1
                            else:
                                actual_status[u] = self.available_statuses['Hospitalized_severe']
                        else:
                            actual_status[u] = self.available_statuses['Hospitalized_mild']
                    self.params['nodes']['tested'][u] = True

                else:
                    recovered = np.random.random_sample()
                    if recovered < self.__get_threshold(u, 'gamma'):
                        actual_status[u] = self.available_statuses['Recovered']
                    else:
                        dead = np.random.random_sample()
                        if dead < self.__get_threshold(u, 'omega'):
                            actual_status[u] = self.available_statuses['Dead']

            ####################### Quarantined Compartment ###########################

            elif u_status == self.available_statuses['Identified_Exposed']:
                at = np.random.random_sample()
                if at < self.__get_threshold(u, 'sigma'):
                    icup = np.random.random_sample()
                    # icu_avalaibility = np.random.random_sample()
                    if icup < self.__get_threshold(u, 'iota'):
                        if self.icu_b > 0 and not self.params['nodes']['ICU'][u]:
                            actual_status[u] = self.available_statuses['Hospitalized_severe_ICU']
                            self.icu_b -= 1
                        else:
                            actual_status[u] = self.available_statuses['Hospitalized_severe']
                    else:
                        actual_status[u] = self.available_statuses['Hospitalized_mild']
                    self.params['nodes']['ICU'][u] = True

            elif u_status == self.available_statuses['Hospitalized_mild']:
                recovered = np.random.random_sample()
                if recovered < self.__get_threshold(u, 'gamma'):
                    actual_status[u] = self.available_statuses['Recovered']
                else:
                    dead = np.random.random_sample()
                    if dead < self.__get_threshold(u, 'omega'):
                        actual_status[u] = self.available_statuses['Dead']

            elif u_status == self.available_statuses['Hospitalized_severe']:
                recovered = np.random.random_sample()
                if recovered < self.__get_threshold(u, 'gamma_f'):
                    actual_status[u] = self.available_statuses['Recovered']
                else:
                    dead = np.random.random_sample()
                    if dead < self.__get_threshold(u, 'omega_f'):
                        actual_status[u] = self.available_statuses['Dead']

            elif u_status == self.available_statuses['Hospitalized_severe_ICU']:
                recovered = np.random.random_sample()
                if recovered < self.__get_threshold(u, 'gamma_t'):
                    actual_status[u] = self.available_statuses['Recovered']
                    self.icu_b += 1
                else:
                    dead = np.random.random_sample()
                    if dead < self.__get_threshold(u, 'omega_t'):
                        actual_status[u] = self.available_statuses['Dead']
                        self.icu_b += 1

            ####################### Lockdown Compartment ###########################

            elif u_status == self.available_statuses['Lockdown_Susceptible']:
                # test lockdown exit
                exit = np.random.random_sample()  # loockdown acceptance
                if exit < self.__get_threshold(u, 'mu'):
                    actual_status[u] = self.available_statuses['Susceptible']
                    self.__ripristinate_social_contacts(u)

                else:
                    actual_status[u] = self.__Susceptible_to_Exposed(u, neighbors, lockdown=True)

            elif u_status == self.available_statuses['Lockdown_Exposed']:
                # test lockdown exit
                exit = np.random.random_sample()  # loockdown exit
                if exit < self.__get_threshold(u, 'mu'):
                    actual_status[u] = self.available_statuses['Exposed']
                    self.__ripristinate_social_contacts(u)

                else:
                    at = np.random.random_sample()
                    if at < self.__get_threshold(u, 'sigma'):
                        actual_status[u] = self.available_statuses['Lockdown_Infected']

            elif u_status == self.available_statuses['Lockdown_Infected']:
                # test lockdown exit
                exit = np.random.random_sample()  # loockdown exit

                if exit < self.__get_threshold(u, 'mu'):
                    actual_status[0] = self.available_statuses['Infected']
                    self.__ripristinate_social_contacts(u)

                else:
                    dead = np.random.random_sample()
                    if dead < self.__get_threshold(u, 'omega'):
                        actual_status[u] = self.available_statuses['Dead']
                    else:
                        recovered = np.random.random_sample()
                        if recovered < self.__get_threshold(u, 'gamma'):
                            actual_status[u] = self.available_statuses['Recovered']

            ####################### Resolved Compartment ###########################

            elif u_status == self.available_statuses['Recovered']:
                immunity = np.random.random_sample()
                if immunity < self.__get_threshold(u, 's'):
                    actual_status[u] = self.available_statuses['Susceptible']

            elif u_status == self.available_statuses['Vaccinated']:
                failure = np.random.random_sample()
                if failure < self.__get_threshold(u, 'f'):
                    self.params['nodes']['vaccinated'][u] = False
                    actual_status[u] = self.available_statuses['Susceptible']

            elif u_status == self.available_statuses['Dead']:
                pass

        delta, node_count, status_delta = self.status_delta(actual_status)
        self.status = actual_status
        self.actual_iteration += 1

        if node_status:
            return {"iteration": self.actual_iteration - 1, "status": delta.copy(),
                    "node_count": node_count.copy(), "status_delta": status_delta.copy()}
        else:
            return {"iteration": self.actual_iteration - 1, "status": {},
                    "node_count": node_count.copy(), "status_delta": status_delta.copy()}

    ###################################################################################################################

    @staticmethod
    def __interaction_selection(neighbors, prob):
        return np.random.choice(a=neighbors, size=int(len(neighbors) * prob), replace=True)

    def add_ICU_beds(self, n):
        """
        Add/Subtract beds in intensive care

        :param n: number of beds to add/remove
        :return:
        """
        self.icu_b = max(0, self.icu_b + n)

    def set_lockdown(self, households=None, workplaces=None):
        """
        Impose the beginning of a lockdown

        :param households: (optional) dictionary specifying the households for each node <node_id -> list(nodes in household)>
        :return:
        """
        actual_status = {node: nstatus for node, nstatus in future.utils.iteritems(self.status)}

        self.lockdown = True

        for u in self.graph.nodes:

            candidate = True
            if workplaces is not None and 'work' in self.params['nodes']:
                if len(set(workplaces) & set(self.params['nodes']['work'][u])) == 0:
                    candidate = False

            if not candidate:
                continue

            la = np.random.random_sample()  # loockdown acceptance
            if la < self.__get_threshold(u, 'lambda'):

                if actual_status[u] == self.available_statuses['Susceptible']:
                    actual_status[u] = self.available_statuses['Lockdown_Susceptible']
                    if households is None or u not in households:
                        self.__limit_social_contacts(u, event="Lockdown")
                    else:
                        self.params['nodes']['filtered'][u] = households[u]

                elif actual_status[u] == self.available_statuses['Exposed']:
                    actual_status[u] = self.available_statuses["Lockdown_Exposed"]
                    if households is None or u not in households:
                        self.__limit_social_contacts(u, event="Lockdown")
                    else:
                        self.params['nodes']['filtered'][u] = households[u]

                elif actual_status[u] == self.available_statuses['Infected']:
                    actual_status[u] = self.available_statuses['Lockdown_Infected']
                    if households is None or u not in households:
                        self.__limit_social_contacts(u, event="Lockdown")
                    else:
                        self.params['nodes']['filtered'][u] = households[u]
            else:
                # node refuses lockdown
                self.params['nodes']['filtered'][u] = None

        delta, node_count, status_delta = self.status_delta(actual_status)
        self.status = actual_status
        return {"iteration": self.actual_iteration - 1, "status": {}, "node_count": node_count.copy(),
                "status_delta": status_delta.copy()}

    def unset_lockdown(self, workplaces=None):
        """
        Remove the lockdown social limitations

        :return:
        """
        actual_status = {node: nstatus for node, nstatus in future.utils.iteritems(self.status)}

        self.lockdown = False

        for u in self.graph.nodes:

            candidate = True
            if workplaces is not None and 'work' in self.params['nodes']:
                if len(set(workplaces) & set(self.params['nodes']['work'][u])) == 0:
                    candidate = False

            if not candidate:
                continue

            self.__ripristinate_social_contacts(u)
            if actual_status[u] == self.available_statuses['Lockdown_Susceptible']:
                actual_status[u] = self.available_statuses['Susceptible']
            elif actual_status[u] == self.available_statuses['Lockdown_Exposed']:
                actual_status[u] = self.available_statuses["Exposed"]
            elif actual_status[u] == self.available_statuses['Lockdown_Infected']:
                actual_status[u] = self.available_statuses['Infected']

        delta, node_count, status_delta = self.status_delta(actual_status)
        self.status = actual_status
        return {"iteration": self.actual_iteration + 1, "status": {}, "node_count": node_count.copy(),
                "status_delta": status_delta.copy()}

    def __limit_social_contacts(self, u, neighbors=None, event='Tested'):

        # already in quarantine or in an household
        if self.params['nodes']['filtered'][u] is None or len(self.params['nodes']['filtered'][u]) > 0:
            return

        if neighbors is None:
            if self.graph.directed:
                neighbors = self.graph.predecessors(u)
            else:
                neighbors = self.graph.neighbors(u)

        if event == 'Tested':
            filtering_prob = self.__get_threshold(u, 'epsilon_e')
            remaining_candidates = [n for n in neighbors if self.params['nodes']['filtered'][n] is not None and len(self.params['nodes']['filtered'][n]) == 0]
            size = min(int(len(neighbors) * (1 - filtering_prob)), len(remaining_candidates))
        else:
            houseold_max_size = self.params['model']['epsilon_l']
            size = list(np.random.choice(a=range(houseold_max_size), size=1))[0]

        # filtering contacts
        to_keep = list(np.random.choice(a=neighbors, size=size, replace=False))
        self.params['nodes']['filtered'][u] = to_keep

        # Lockdown
        if event != "Tested":
            for n in to_keep:
                # if n complies to the lockdown and is not in quarantine
                if self.params['nodes']['filtered'][n] is not None and \
                         self.status[n] not in [self.available_statuses['Identified_Exposed'],
                                                self.available_statuses['Hospitalized_mild'],
                                                self.available_statuses['Hospitalized_severe_ICU'],
                                                self.available_statuses['Hospitalized_severe']]:

                    # imposing an estimate "household" lockdown
                    self.params['nodes']['filtered'][n] = list((set(to_keep) - {n}) | {u})
        else:
            # Quarantine
            for n in set(neighbors) - set(to_keep):
                if self.params['nodes']['filtered'][n] is not None and len(self.params['nodes']['filtered'][n]) > 0:
                    # removing the quarantined node from the neighbors' contacts
                    self.params['nodes']['filtered'][n] = list(set(self.params['nodes']['filtered'][n]) - {u})
                else:
                    if self.status[n] not in [self.available_statuses['Identified_Exposed'],
                                              self.available_statuses['Hospitalized_mild'],
                                              self.available_statuses['Hospitalized_severe_ICU'],
                                              self.available_statuses['Hospitalized_severe']]:

                        if self.graph.directed:
                            n_neighbors = self.graph.predecessors(u)
                        else:
                            n_neighbors = self.graph.neighbors(u)
                        self.params['nodes']['filtered'][n] = list(set(n_neighbors) - {u})

    def __ripristinate_social_contacts(self, u):
        self.params['nodes']['filtered'][u] = []

    ####################### Undetected Compartment ###########################

    def __Susceptible_to_Exposed(self, u, neighbors, lockdown=False):

        # vaccination test
        if self.__get_threshold(u, 'v') > 0 and not self.params['nodes']['vaccinated'][u]:
            v_prob = np.random.random_sample()
            if v_prob < self.__get_threshold(u, 'v'):
                self.params['nodes']['vaccinated'][u] = True
                return self.available_statuses['Vaccinated']

        interactions = list(self.__interaction_selection(neighbors, self.params['nodes']['activity'][u]))
        social_interactions = len(interactions)

        l_range_proba = self.__get_threshold(u, 'p')
        if lockdown:
            l_range_proba = self.__get_threshold(u, 'p_l')

        l_range = np.random.random_sample()  # long range interaction
        if l_range < l_range_proba:
            # filtering out quarantined and dead nodes
            if self.__get_threshold(u, 'z') == 0:
                candidates = [n for n in self.graph.nodes if self.status[n] not in
                              [self.available_statuses['Identified_Exposed'],
                               self.available_statuses['Hospitalized_mild'],
                               self.available_statuses['Dead']]]
            else:
                candidates = [n for n in self.graph.nodes if self.status[n] not in
                              [self.available_statuses['Identified_Exposed'],
                               self.available_statuses['Hospitalized_mild']]
                              ]

            interactions.extend(list(np.random.choice(a=candidates,
                                                      size=int(social_interactions * self.params['model']['lsize']),
                                                      replace=True)))

        for v in interactions:
            if self.status[v] == self.available_statuses['Infected'] or \
                    self.status[v] == self.available_statuses['Lockdown_Infected'] or \
                    self.status[v] == self.available_statuses['Hospitalized_mild']:
                bt = np.random.random_sample()

                if bt < self.__get_threshold(u, 'beta'):
                    if lockdown:
                        return self.available_statuses['Lockdown_Exposed']
                    return self.available_statuses['Exposed']

            elif self.status[v] == self.available_statuses['Dead']:
                zp = np.random.random_sample()
                if zp < self.__get_threshold(u, 'z'):  # infection risk due to partial corpse disposal
                    if lockdown:
                        return self.available_statuses['Lockdown_Exposed']
                    return self.available_statuses['Exposed']

        if lockdown:
            return self.available_statuses['Lockdown_Susceptible']
        return self.available_statuses['Susceptible']

    def __get_threshold(self, u, parameter):
        # stratified population scenario
        if isinstance(self.params['model'][parameter], dict):
            n_class = self.params['nodes']['segment'][u]
            return self.params['model'][parameter][n_class]
        # base scenario, single value
        else:
            return self.params['model'][parameter]




# TODO
# - Fix visualization logic (overwrite vs update),
# - numpy matrix implementation instead of networkx nodes/dict
# Requirements, networkx, numpy, matplotlib, PIL, pyintergraph, python-igraph, tqdm

from ndlib.models.DiffusionModel import DiffusionModel
import future.utils
import os
import networkx as nx
import copy
import numpy as np

import matplotlib.pyplot as plt
import matplotlib as mpl
import matplotlib.animation as animation

__author__ = 'Mathijs Maijer'
__license__ = "BSD-2-Clause"
__email__ = "m.f.maijer@gmail.com"


class ContinuousModel(DiffusionModel):

    def __init__(self, graph, constants=None, clean_status=None, iteration_schemes=None, save_file=None):
        """
        Model Constructor

        :param graph: A networkx graph object
        :param constants: dictionary containing state name as key and float or function returning a float as value
        :param clean_status: boolean indicating whether to set all status values between 0 and 1
        :param iteration_schemes: list of dictionaries for each scheme
            containing a name, function(graph, status), and optional lower and upper bound keys
        :param save_file: string indicating path and file name to save the iterations output to
         """
        super(ContinuousModel, self).__init__(graph)
        self.compartment = {}
        self.compartment_progressive = 0
        self.status_progressive = 0

        self.discrete_state = False

        self.available_statuses = {
            "Infected": 0
        }

        self.clean = True if clean_status else False

        self.constants = constants

        if iteration_schemes:
            self.iteration_schemes = iteration_schemes
            self.iteration_schemes.append({'name': '', 'function': lambda graph, status: graph.nodes})
        else:
            self.iteration_schemes = [{'name': '', 'function': lambda graph, status: graph.nodes}]

        self.visualization_configuration = None

        if save_file:
            if isinstance(save_file, str):
                self.save_file = save_file
            else:
                raise ValueError('save_file should be a string indicating path/and/filename')
        else:
            self.save_file = None

        self.full_status = None

    def configure_visualization(self, visualization_configuration):
        """
        Configure and assert all visualization configuration parameters

        :param visualization_configuration: dictionary containing all visualization options
        """
        if visualization_configuration:
            print('Configuring visualization...')
            self.visualization_configuration = visualization_configuration
            vis_keys = visualization_configuration.keys()

            self.validate_plot_config(visualization_configuration, vis_keys)
            self.validate_color_config(vis_keys)

            if 'pos' not in self.graph.nodes[0].keys():
                self.configure_layout(vis_keys)

            if 'variable_limits' not in vis_keys:
                self.visualization_configuration['variable_limits'] = {key: [-1, 1] for key in list(self.available_statuses.keys())}
            else:
                for key in list(self.available_statuses.keys()):
                    if key not in list(self.visualization_configuration['variable_limits'].keys()):
                        self.visualization_configuration['variable_limits'][key] = [-1, 1]
            if 'animation_interval' not in vis_keys:
                self.visualization_configuration['animation_interval'] = 30
            else:
                if not isinstance(self.visualization_configuration['animation_interval'], int):
                    raise ValueError('animation interval must be an integer')
            print('Done configuring the visualization')
        else:
            raise Exception('Provide a visualization configuration when using this function')

    def configure_layout(self, vis_keys):
        if 'layout' in vis_keys:
            if self.visualization_configuration['layout'] == 'fr':
                import pyintergraph
                Graph = pyintergraph.InterGraph.from_networkx(self.graph.graph)
                G = Graph.to_igraph()
                layout = G.layout_fruchterman_reingold(niter=500)
                positions = {node: {'pos': location} for node, location in enumerate(layout)}
            else:
                if 'layout_params' in vis_keys:
                    pos = self.visualization_configuration['layout'](self.graph.graph, **self.visualization_configuration['layout_params'])
                else:
                    pos = self.visualization_configuration['layout'](self.graph.graph)
                positions = {key: {'pos': location} for key, location in pos.items()}
        else:
            pos = nx.drawing.spring_layout(self.graph.graph)
            positions = {key: {'pos': location} for key, location in pos.items()}

        nx.set_node_attributes(self.graph, positions)

    def validate_plot_config(self, visualization_configuration, vis_keys):
        if 'plot_interval' in vis_keys:
            if isinstance(visualization_configuration['plot_interval'], int):
                if visualization_configuration['plot_interval'] <= 0:
                    raise ValueError('plot_interval must be a positive integer')
            else:
                raise ValueError('plot_interval must be a positive integer')
        else:
            raise ValueError('plot_interval must be included for visualization')
        if 'show_plot' in vis_keys:
            if not isinstance(visualization_configuration['show_plot'], bool):
                raise ValueError('show_plot must be a boolean')
        else:
            self.visualization_configuration['show_plot'] = True
        if 'plot_variable' in vis_keys:
            if not isinstance(visualization_configuration['plot_variable'], str):
                raise ValueError('Plot variable must be a string')
        else:
            self.visualization_configuration['plot_variable'] = None

        if 'plot_title' in self.visualization_configuration.keys():
            if not isinstance(self.visualization_configuration['plot_title'], str):
                raise ValueError('Plot name must be a string')
        else:
            vis_var = self.visualization_configuration['plot_variable']
            self.visualization_configuration['plot_title'] = 'Network simulation of ' + vis_var

        if 'plot_annotation' in vis_keys:
            if not isinstance(self.visualization_configuration['plot_annotation'], str):
                raise ValueError('Plot annotation must be a string')
        else:
            self.visualization_configuration['plot_annotation'] = None

    def validate_color_config(self, vis_keys):
        if 'cmin' in vis_keys:
            if not isinstance(self.visualization_configuration['cmin'], int):
                raise ValueError('cmin must be an integer')
        else:
            self.visualization_configuration['cmin'] = 0

        if 'cmax' in vis_keys:
            if not isinstance(self.visualization_configuration['cmax'], int):
                raise ValueError('cmax must be an integer')
        else:
            self.visualization_configuration['cmax'] = 1

        if 'color_scale' in vis_keys:
            if not isinstance(self.visualization_configuration['color_scale'], str):
                raise ValueError('Color scale must be a string')
        else:
            self.visualization_configuration['color_scale'] = 'RdBu'

    def add_status(self, status_name):
        """
        Add a status/state to the model
        """
        if status_name not in self.available_statuses:
            self.available_statuses[status_name] = self.status_progressive
            self.status_progressive += 1

    def add_rule(self, status, function, rule, schemes=['']):
        """
        Add a rule to the model

        :param status: string indicating the status
        :param function: A function that updates the status value
            it receives the parameters: node, graph, status, attributes, constants
        :param rule: A condition that should be true before the status is updated
        :param schemes: A list of strings matching the names of the schemes in which the rule should be assessed
            If no schemes are provided, the default scheme '' is used, which is assessed for every iteration
        """
        self.compartment[self.compartment_progressive] = (status, function, rule, schemes)
        self.compartment_progressive += 1

    def set_initial_status(self, initial_status_funs, configuration=None):
        """
        Override behaviour of methods in class DiffusionModel.
        Overwrites initial status using given function
        Generates node profiles
        """
        super(ContinuousModel, self).set_initial_status(configuration)

        if not isinstance(initial_status_funs, dict):
            raise ValueError('The initial status should be a dictionary of form status (str): value (int/float/function)')

        # set node status
        for node in self.status:
            status = {}
            for status_fun in initial_status_funs.items():
                if hasattr(status_fun[1], '__call__'):
                    status[status_fun[0]] = status_fun[1](node, self.graph, status, self.constants)
                    continue
                if not isinstance(status_fun[1], float):
                    if not isinstance(status_fun[1], int):
                        raise ValueError('The initial status should be a function, integer, or float')
                status[status_fun[0]] = status_fun[1]
            self.status[node] = status

        self.initial_status = copy.deepcopy(self.status)

    def clean_initial_status(self, valid_status=None):
        """
        For every status, set it to 0 if negative, or to 1 if > 1
        """
        for n, s in future.utils.iteritems(self.status):
            for var_val in s.items():
                if var_val[1] > 1:
                    self.status[n][var_val[0]] = 1
                elif var_val[1] < 0:
                    self.status[n][var_val[0]] = 0

    def iteration(self, node_status=True):
        """
        Execute a single model iteration

        :return: Iteration_id, Incremental node status (dictionary node->status)
        """
        if self.clean:
            self.clean_initial_status()

        # actual_status = {node: nstatus for node, nstatus in future.utils.iteritems(self.status)}
        actual_status = copy.deepcopy(self.status)

        if self.actual_iteration == 0:
            self.actual_iteration += 1
            delta, status_delta = self.status_delta_continuous(self.status)
            if node_status:
                return {"iteration": 0, "status": copy.deepcopy(self.status),
                        "status_delta": copy.deepcopy(status_delta)}
            else:
                return {"iteration": 0, "status": {},
                         "status_delta": copy.deepcopy(status_delta)}

        nodes_data = self.graph.nodes

        for scheme in self.iteration_schemes:

            if 'lower' in scheme and scheme['lower'] > self.actual_iteration:
                continue
            if 'upper' in scheme and scheme['upper'] < self.actual_iteration:
                continue

            nodes = scheme['function'](self.graph, self.status)
            for u in nodes:
                # For all rules
                for i in range(0, self.compartment_progressive):
                    if scheme['name'] in self.compartment[i][3]:
                        # Get and test the condition
                        rule = self.compartment[i][2]
                        test = rule.execute(node=u, graph=self.graph, status=self.status,
                                            status_map=self.available_statuses, attributes=nodes_data,
                                            params=self.params, constants=self.constants)
                        if test:
                            # Update status or network if test succeeds
                            if self.compartment[i][0] == 'network':
                                self.compartment[i][1](u, self.graph, self.status, nodes_data, self.constants)
                            else:
                                val = self.compartment[i][1](u, self.graph, self.status, nodes_data, self.constants)
                                actual_status[u][self.compartment[i][0]] = val

        delta, status_delta = self.status_delta_continuous(actual_status)
        self.status = actual_status
        self.actual_iteration += 1

        if node_status:
            return {"iteration": self.actual_iteration - 1, "status": copy.deepcopy(delta),
                    "status_delta": copy.deepcopy(status_delta)}
        else:
            return {"iteration": self.actual_iteration - 1, "status": {},
                    "status_delta": copy.deepcopy(status_delta)}

    def iteration_bunch(self, bunch_size, node_status=True, tqdm=True):
        """
        Execute bunch_size of model iterations and save the result if save_file is set

        :param bunch_size: integer number of iterations to execute
        :param node_status: boolean indicating whether to keep the statuses of the nodes
        :param tqdm: boolean indicating whether to use tqdm to show the estimated duration

        :return: list of outputs for every iteration
        """
        if tqdm:
            iterations = super().iteration_bunch(bunch_size, node_status)
        else:
            iterations = []
            for _ in range(bunch_size):
                iterations.append(self.iteration(node_status))

        if self.save_file:
            split = self.save_file.split('/')
            file_name = split[-1]
            file_path = self.save_file.replace(file_name, '')
            if not os.path.exists(file_path):
                os.makedirs(file_path)
            np.save(self.save_file, iterations)
            print('Saved ' + self.save_file)
        return iterations

    def get_mean_data(self, iterations, mean_type):
        """
        Create a dictionary with statuses as keys, and a list of average value per iteration as value

        :param iterations: iterations output from iteration_bunch
        :param mean_type: A string containing the type to get the average data from
            Should be 'status' or 'status_delta'

        :return: Dictionary containing all statuses as keys,
            and as value a list of the average values of all nodes for that status per iteration
        """
        mean_changes = {}
        for key in self.available_statuses.keys():
            mean_changes[key] = []
        del mean_changes['Infected'] # Todo, fix this line so that infected isn't even in available statuses

        for it in iterations:
            delta = {}

            vals = list(it[mean_type].values())

            for val in vals:
                for key, v in val.items():
                    if key not in delta.keys():
                        delta[key] = {'v': v, 'n': 1}
                    else:
                        delta[key]['v'] += v
                        delta[key]['n'] += 1
            for key in mean_changes.keys():
                if key not in delta.keys():
                    delta[key] = {}
                    delta[key]['v'] = 0
                    delta[key]['n'] = 1

            for k, v in delta.items():
                if k not in mean_changes.keys():
                    mean_changes[k] = []
                mean_changes[k].append(delta[k]['v'] / delta[k]['n'])

        return mean_changes

    def build_full_status(self, iterations):
        """
        Create a list with objects that have all the nodes with their statuses per iteration

        :param iterations: iterations output from iteration_bunch

        :return: a list of status objects that contain an iteration key with its number as value and a status key
            the status value contains per node all its states as keys with the corresponding state values
        """
        statuses = []
        status = {'iteration': 0, 'status': {}}
        for key, val in iterations[0]['status'].items():
            status['status'][key] = val
        statuses.append(status)
        for it in iterations[1:]:
            i = it['iteration']
            status = copy.deepcopy(statuses[-1])
            for node, d in it['status'].items():
                for var, val in d.items():
                    status['status'][node][var] = val
                    status['iteration'] = i
            statuses.append(status)

        return statuses

    def get_means(self, iterations):
        """
        Create a full status and get the mean data for the status key

        :param iterations: iterations output from iteration_bunch

        :return: Dictionary containing all statuses as keys,
            and as value a list of the average value over all nodes for that state per iteration
        """
        self.full_status = self.build_full_status(iterations)
        means = self.get_mean_data(self.full_status, 'status')
        return means

    def build_trends(self, iterations):
        """
        Overwrite build trends of diffusionmodel
        """
        means = self.get_means(iterations)
        means_status_delta_vals = self.get_mean_data(iterations, 'status')
        status_delta = self.get_mean_data(iterations, 'status_delta')

        return {'mean_delta_status_vals': means_status_delta_vals, 'status_delta': status_delta, 'means': means}

    def plot(self, trends, n, delta=None, delta_mean=None):
        """
        Create and show different plots of the trends

        :param trends: output from the build_trends function
        :param n: integer amount of iterations to show
        :param delta: boolean indicating whether to show the mean change per variable per iteration
        :param delta_mean: boolean indicating whether to show the mean value of changed variables per iteration
        """
        x = np.arange(0, n)

        sub_plots = 1
        if delta:
            sub_plots = 2 if (delta and not delta_mean) else 3

        _, axs = plt.subplots(sub_plots)

        # Mean status delta per iterations
        if delta or delta_mean:
            for status, values in trends['means'].items():
                axs[0].plot(x, values, label=status)
            axs[0].set_title("Mean values per variable per iteration")
            axs[0].legend()
        else:
            for status, values in trends['means'].items():
                plt.plot(x, values, label=status)
            plt.title("Mean values per variable per iteration")
            plt.legend()


        if delta:
            for status, values in trends['status_delta'].items():
                axs[1].plot(x, values, label=status)
            axs[1].set_title("Mean change per variable per iteration")
            axs[1].legend()

        i = 2 if delta else 1

        if delta_mean:
            for status, values in trends['mean_delta_status_vals'].items():
                axs[i].plot(x, values, label=status)
            axs[i].set_title("Mean value of changed variables per iteration")
            axs[i].legend()

        plt.show()

    def create_frames(self, iterations):
        """
        Create frames every plot_interval iterations
            by creating a dictionary that contains a list of all node values for a status key per iteration

        :param iterations: iterations output from iteration_bunch

        :return: tuple of a status value dictionary and a list of lists of node colors,
            the dictionary has a status as key and as value a list of lists, the first dimension corresponds with the iteration,
            the second dimension is a list that contains all the different node status values for that iteration,
            the node_colors list first dimension corresponds with the iteration,
            and the second dimension is a list of lists where each index holds the color value of the node at that index
        """
        if not self.full_status:
            self.full_status = self.build_full_status(iterations)
        statuses = list(self.available_statuses.keys())
        statuses.remove('Infected')
        histo_frames = {key: [] for key in statuses}
        node_colors = []
        for i in range(len(self.full_status)):
            if i % self.visualization_configuration['plot_interval'] == 0:
                status_list = {key: [] for key in statuses}
                for node in self.full_status[i]['status'].keys():
                    vals = self.full_status[i]['status'][node]
                    for key in statuses:
                        status_list[key].append(vals[key])

                for key in statuses:
                    histo_frames[key].append(status_list[key])
                node_colors.append([self.full_status[i]['status'][node][self.visualization_configuration['plot_variable']] for node in self.graph.nodes])

        return (histo_frames, node_colors)

    def visualize(self, iterations):
        """
        Visualize the network and color the nodes using a status
        Show histograms of the other statuses beneath the graph
        All visualization options are specified in the visualization_configuration variable
        If show_plot is set to True in the configuration, the visualization will be shown dynamically
        If plot_output is set in the configuration, an animation will be saved

        Currently the graphs are cleared and then completely redrawn, a good optimization would be to only update the changed values

        :param iterations: iterations output from iteration_bunch
        """
        if not self.visualization_configuration:
            raise Exception("Specify a visualization configuration before you visualize the model")

        (histo_frames, node_colors) = self.create_frames(iterations)

        statuses = list(self.available_statuses.keys())
        statuses.remove('Infected')

        n_status = len(statuses)

        fig = plt.figure(figsize=(10,9), constrained_layout=True)
        gs = fig.add_gridspec(6, n_status)

        network = fig.add_subplot(gs[:-1, :])

        axis = []

        for i in range(n_status):
            ax = fig.add_subplot(gs[-1, i])
            ax.set_title(statuses[i])
            ax.get_xaxis().set_ticks([])
            ax.get_yaxis().set_ticks([])
            axis.append(ax)

        n = int(len(iterations)/self.visualization_configuration['plot_interval'])

        cm = plt.cm.get_cmap(self.visualization_configuration['color_scale'])
        vmin = self.visualization_configuration['variable_limits'][self.visualization_configuration['plot_variable']][0]
        vmax = self.visualization_configuration['variable_limits'][self.visualization_configuration['plot_variable']][1]

        def updateData(curr):
            # Clean previous graphs
            network.clear()
            for ax in axis:
                ax.clear()

            # Plot all variable histograms
            for i, ax in enumerate(axis):
                _, bins, patches = ax.hist(histo_frames[statuses[i]][curr], range=self.visualization_configuration['variable_limits'][statuses[i]], density=1, bins=25, edgecolor='black')
                bin_centers = 0.5 * (bins[:-1] + bins[1:])
                col = bin_centers - min(bin_centers)
                col /= max(col)
                for c, p in zip(col, patches):
                    plt.setp(p, 'facecolor', cm(c))
                ax.set_title(statuses[i])
                # ax.set_ylim([0, len(self.graph.nodes)])
                ax.get_xaxis().set_ticks([])
                ax.get_yaxis().set_ticks([])

            # Plot network
            pos = nx.get_node_attributes(self.graph.graph, 'pos')
            nx.draw_networkx_edges(self.graph.graph, pos, alpha=0.2, ax=network)
            nc = nx.draw_networkx_nodes(self.graph.graph, pos, nodelist=self.graph.nodes, node_color=node_colors[curr], vmin=vmin, vmax=vmax, cmap=cm, node_size=50, ax=network)
            nc.set_edgecolor('black')
            network.get_xaxis().set_ticks([])
            network.get_yaxis().set_ticks([])
            network.set_title('Iteration: ' + str(curr * self.visualization_configuration['plot_interval']))

        simulation = animation.FuncAnimation(fig, updateData, n, interval=self.visualization_configuration['animation_interval'], repeat=True)

        norm = mpl.colors.Normalize(vmin=vmin, vmax=vmax)
        sm = plt.cm.ScalarMappable(cmap=cm, norm=norm)
        sm.set_array([])
        fig.colorbar(sm, ax=network)
        fig.suptitle(self.visualization_configuration['plot_title'], fontsize=16)

        if self.visualization_configuration['show_plot']:
            plt.show()

        if 'plot_output' in self.visualization_configuration.keys():
            self.save_plot(simulation)

    def save_plot(self, simulation):
        """
        Save the plot to a file specified in plot_output int he visualization configuration
        The file is generated using the writer from the pillow library

        :param simulation: Output of the matplotlib animation.FuncAnimation function
        """
        print('Saving plot at: ' + self.visualization_configuration['plot_output'] + ' ...')
        split = self.visualization_configuration['plot_output'].split('/')
        file_name = split[-1]
        file_path = self.visualization_configuration['plot_output'].replace(file_name, '')
        if not os.path.exists(file_path):
            os.makedirs(file_path)

        from PIL import Image
        writergif = animation.PillowWriter(fps=5)
        simulation.save(self.visualization_configuration['plot_output'], writer=writergif)
        print('Saved: ' + self.visualization_configuration['plot_output'])



from ..DiffusionModel import DiffusionModel
import numpy as np
import future

__author__ = ["Giulio Rossetti"]
__license__ = "BSD-2-Clause"


class SEIRctModel(DiffusionModel):

    def __init__(self, graph, seed=None):

        super(self.__class__, self).__init__(graph, seed)

        self.name = "SEIR"

        self.available_statuses = {
            "Susceptible": 0,
            "Exposed": 2,
            "Infected": 1,
            "Removed": 3
        }
        self.parameters = {
            "model": {
                "alpha": {
                    "descr": "Latent period (1/duration)",
                    "range": [0, 1],
                    "optional": False},
                "beta": {
                    "descr": "Infection rate",
                    "range": [0, 1],
                    "optional": False},
                "gamma": {
                    "descr": "Recovery rate",
                    "range": [0, 1],
                    "optional": False
                },
                "tp_rate": {
                    "descr": "Whether if the infection rate depends on the number of infected neighbors",
                    "range": [0, 1],
                    "optional": True,
                    "default": 1
                }
            },
            "nodes": {},
            "edges": {},
        }

        self.progress = {}
        self.progress_I = {}

    def iteration(self, node_status=True):
        self.clean_initial_status(self.available_statuses.values())

        actual_status = {node: nstatus for node, nstatus in future.utils.iteritems(self.status)}
        self.progress_I = {node: 0 for node in actual_status if actual_status[node] == 1}

        if self.actual_iteration == 0:
            self.actual_iteration += 1
            delta, node_count, status_delta = self.status_delta(actual_status)
            if node_status:
                return {"iteration": 0, "status": actual_status.copy(),
                        "node_count": node_count.copy(), "status_delta": status_delta.copy()}
            else:
                return {"iteration": 0, "status": {},
                        "node_count": node_count.copy(), "status_delta": status_delta.copy()}

        for u in self.graph.nodes:

            u_status = self.status[u]
            eventp = np.random.random_sample()
            neighbors = self.graph.neighbors(u)
            if self.graph.directed:
                neighbors = self.graph.predecessors(u)

            if u_status == 0:  # Susceptible

                infected_neighbors = [v for v in neighbors if self.status[v] == 1]
                triggered = 1 if len(infected_neighbors) > 0 else 0

                if self.params['model']['tp_rate'] == 1:
                    if eventp < 1 - (1 - self.params['model']['beta']) ** len(infected_neighbors):
                        actual_status[u] = 2  # Exposed
                        self.progress[u] = self.actual_iteration  # save time of exposure t_i
                else:
                    if eventp < self.params['model']['beta'] * triggered:
                        actual_status[u] = 2  # Exposed
                        self.progress[u] = self.actual_iteration  # save time of exposure t_i

            elif u_status == 2:

                # apply prob. of infection, after (t - t_i) 
                if eventp < 1 - np.exp(- (self.actual_iteration - self.progress[u]) * self.params['model']['alpha']): 
                    actual_status[u] = 1  # Infected
                    self.progress_I[u] = self.actual_iteration
                    del self.progress[u]

            elif u_status == 1:
                if eventp < 1 - np.exp(- (self.actual_iteration - self.progress_I[u]) * self.params['model']['gamma']):
                    actual_status[u] = 3  # Removed
                    del self.progress_I[u]

        delta, node_count, status_delta = self.status_delta(actual_status)
        self.status = actual_status
        self.actual_iteration += 1

        if node_status:
            return {"iteration": self.actual_iteration - 1, "status": delta.copy(),
                    "node_count": node_count.copy(), "status_delta": status_delta.copy()}
        else:
            return {"iteration": self.actual_iteration - 1, "status": {},
                    "node_count": node_count.copy(), "status_delta": status_delta.copy()}



from ..DynamicDiffusionModel import DynamicDiffusionModel
import numpy as np
import networkx as nx
import future.utils
from scipy import stats

__author__ = "Letizia Milli"
__license__ = "BSD-2-Clause"
__email__ = "letizia.milli@isti.cnr.it"


class DynProfileThresholdModel(DynamicDiffusionModel):
    """
        Node Parameters to be specified via ModelConfig

        :param profile: The node profile. As default a value of 0.1 is assumed for all nodes.
        :param threshold: The node threshold. As default a value of 0.1 is assumed for all nodes.
    """

    def __init__(self, graph, seed=None):
        """
             Model Constructor

             :param graph: A networkx graph object
         """
        super(self.__class__, self).__init__(graph, seed)
        self.available_statuses = {
            "Susceptible": 0,
            "Infected": 1,
            "Blocked": -1
        }

        self.parameters = {
            "model": {
                "blocked": {
                    "descr": "Presence of blocked nodes",
                    "range": [0, 1],
                    "optional": True,
                    "default": 0
                },
                "adopter_rate": {
                    "descr": "Exogenous adoption rate",
                    "range": [0, 1],
                    "optional": True,
                    "default": 0
                }
            },
            "nodes": {
                "threshold": {
                    "descr": "Node threshold",
                    "range": [0, 1],
                    "optional": True,
                    "default": 0.1
                },
                "profile": {
                    "descr": "Node profile",
                    "range": [0, 1],
                    "optional": True,
                    "default": 0.1
                }
            },
            "edges": {},
        }

        self.name = "Profile-Threshold"

    def iteration(self, node_status=True):
        """
        Execute a single model iteration

        :return: Iteration_id, Incremental node status (dictionary node->status)
        """
        self.clean_initial_status(self.available_statuses.values())
        actual_status = {node: nstatus for node, nstatus in future.utils.iteritems(self.status)}

        # streaming
        if self.stream_execution:
            raise ValueError("Streaming network not allowed.")
        # snapshot
        else:
            if self.actual_iteration == 0:
                self.actual_iteration += 1
                delta, node_count, status_delta = self.status_delta(actual_status)
                if node_status:
                    return {"iteration": 0, "status": actual_status.copy(),
                            "node_count": node_count.copy(), "status_delta": status_delta.copy()}
                else:
                    return {"iteration": 0, "status": {},
                            "node_count": node_count.copy(), "status_delta": status_delta.copy()}

            for u in self.graph.nodes():
                if actual_status[u] != 0:
                    continue

                if self.params['model']['adopter_rate'] > 0:
                    xk = (0, 1)
                    pk = (1 - self.params['model']['adopter_rate'], self.params['model']['adopter_rate'])
                    probability = stats.rv_discrete(name='probability', values=(xk, pk))
                    number_probability = probability.rvs()

                    if number_probability == 1:
                        actual_status[u] = 1
                        continue

                neighbors = list(self.graph.neighbors(u))
                if isinstance(self.graph, nx.DiGraph):
                    neighbors = list(self.graph.predecessors(u))

                infected = 0
                for v in neighbors:
                    infected += self.status[v]

                if infected > 0 and actual_status[u] == 0:

                    infected_ratio = float(infected) / len(neighbors)
                    if infected_ratio >= self.params['nodes']['threshold'][u]:
                        eventp = np.random.random_sample()
                        if eventp >= self.params['nodes']['profile'][u]:
                            actual_status[u] = 1
                        else:
                            if self.params['model']['blocked'] != 0:
                                blip = np.random.random_sample()
                                if blip > self.params['model']['blocked']:
                                    actual_status[u] = -1

            delta, node_count, status_delta = self.status_delta(actual_status)
            self.status = actual_status
            self.actual_iteration += 1

            if node_status:
                return {"iteration": self.actual_iteration - 1, "status": delta.copy(),
                        "node_count": node_count.copy(), "status_delta": status_delta.copy()}
            else:
                return {"iteration": self.actual_iteration - 1, "status": {},
                        "node_count": node_count.copy(), "status_delta": status_delta.copy()}



from ..DynamicDiffusionModel import DynamicDiffusionModel
import numpy as np
import networkx as nx
import future.utils

__author__ = "Giulio Rossetti"
__license__ = "BSD-2-Clause"
__email__ = "giulio.rossetti@gmail.com"


class DynSISModel(DynamicDiffusionModel):
    """
       Model Parameters to be specified via ModelConfig

       :param beta: The infection rate (float value in [0,1])
       :param lambda: The recovery rate (float value in [0,1])
    """

    def __init__(self, graph, seed=None):
        """
             Model Constructor

             :param graph: A networkx graph object
         """
        super(self.__class__, self).__init__(graph, seed)
        self.available_statuses = {
            "Susceptible": 0,
            "Infected": 1
        }

        self.parameters = {
            "model": {
                "beta": {
                    "descr": "Infection rate",
                    "range": [0, 1],
                    "optional": False},
                "lambda": {
                    "descr": "Recovery rate",
                    "range": [0, 1],
                    "optional": False
                }
            },
            "nodes": {},
            "edges": {},
        }

        self.name = "SIS"

    def iteration(self, node_status=True):
        """
        Execute a single model iteration

        :return: Iteration_id, Incremental node status (dictionary node->status)
        """
        self.clean_initial_status(self.available_statuses.values())

        actual_status = {node: nstatus for node, nstatus in future.utils.iteritems(self.status)}

        # streaming
        if self.stream_execution:
            u, v = list(self.graph.edges())[0]
            u_status = self.status[u]
            v_status = self.status[v]

            # infection test
            if u_status == 1 and v_status == 0:
                p = np.random.random_sample()
                if p < self.params['model']['beta']:
                    actual_status[v] = 1

            if v_status == 1 and u_status == 0:
                p = np.random.random_sample()
                if p < self.params['model']['beta']:
                    actual_status[u] = 1

            # removal test
            if v_status == 1:
                g = np.random.random_sample()
                if g < self.params['model']['lambda']:
                    actual_status[v] = 0

            if u_status == 1:
                g = np.random.random_sample()
                if g < self.params['model']['lambda']:
                    actual_status[u] = 0

            delta, node_count, status_delta = self.status_delta(actual_status)
            self.status = actual_status
            self.actual_iteration += 1

            if node_status:
                return {"iteration": self.actual_iteration - 1, "status": delta.copy(),
                        "node_count": node_count.copy(), "status_delta": status_delta.copy()}
            else:
                return {"iteration": self.actual_iteration - 1, "status": {},
                        "node_count": node_count.copy(), "status_delta": status_delta.copy()}
        # snapshot
        else:

            if self.actual_iteration == 0:
                self.actual_iteration += 1
                delta, node_count, status_delta = self.status_delta(actual_status)
                if node_status:
                    return {"iteration": 0, "status": actual_status.copy(),
                            "node_count": node_count.copy(), "status_delta": status_delta.copy()}
                else:
                    return {"iteration": 0, "status": {},
                            "node_count": node_count.copy(), "status_delta": status_delta.copy()}

            for u in self.graph.nodes():

                u_status = self.status[u]
                eventp = np.random.random_sample()
                neighbors = self.graph.neighbors(u)
                if isinstance(self.graph, nx.DiGraph):
                    neighbors = self.graph.predecessors(u)

                if u_status == 0:
                    infected_neighbors = len([v for v in neighbors if self.status[v] == 1])
                    if eventp < self.params['model']['beta'] * infected_neighbors:
                        actual_status[u] = 1
                elif u_status == 1:
                    if eventp < self.params['model']['lambda']:
                        actual_status[u] = 0

        delta, node_count, status_delta = self.status_delta(actual_status)
        self.status = actual_status
        self.actual_iteration += 1

        if node_status:
            return {"iteration": self.actual_iteration - 1, "status": delta.copy(),
                    "node_count": node_count.copy(), "status_delta": status_delta.copy()}
        else:
            return {"iteration": self.actual_iteration - 1, "status": {},
                    "node_count": node_count.copy(), "status_delta": status_delta.copy()}




from ..DiffusionModel import DiffusionModel
import numpy as np
import future.utils

__author__ = "Elisa Salatti"
__license__ = "BSD-2-Clause"


class SEISctModel(DiffusionModel):
    """
       Model Parameters to be specified via ModelConfig

       :param beta: The infection rate (float value in [0,1])
       :param lambda: The recovery rate (float value in [0,1])
    """

    def __init__(self, graph, seed=None):
        """
             Model Constructor

             :param graph: A networkx graph object
         """
        super(self.__class__, self).__init__(graph, seed)
        self.available_statuses = {
            "Susceptible": 0,
            "Exposed": 2,
            "Infected": 1
        }

        self.parameters = {
            "model": {
                "alpha": {
                    "descr": "Incubation period",
                    "range": [0, 1],
                    "optional": False},
                "beta": {
                    "descr": "Infection rate",
                    "range": [0, 1],
                    "optional": False},
                "lambda": {
                    "descr": "Recovery rate",
                    "range": [0, 1],
                    "optional": False
                },
                "tp_rate": {
                    "descr": "Whether if the infection rate depends on the number of infected neighbors",
                    "range": [0, 1],
                    "optional": True,
                    "default": 1
                }
            },
            "nodes": {},
            "edges": {},
        }

        self.name = "SEIS"
        self.progress = {}
        self.progress_I = {}

    def iteration(self, node_status=True):
        """
        Execute a single model iteration

        :return: Iteration_id, Incremental node status (dictionary node->status)
        """
        self.clean_initial_status(self.available_statuses.values())

        actual_status = {node: nstatus for node, nstatus in future.utils.iteritems(self.status)}
        self.progress_I = {node: 0 for node in actual_status if actual_status[node] == 1}

        if self.actual_iteration == 0:
            self.actual_iteration += 1
            delta, node_count, status_delta = self.status_delta(actual_status)
            if node_status:
                return {"iteration": 0, "status": actual_status.copy(),
                        "node_count": node_count.copy(), "status_delta": status_delta.copy()}
            else:
                return {"iteration": 0, "status": {},
                        "node_count": node_count.copy(), "status_delta": status_delta.copy()}

        for u in self.graph.nodes:

            u_status = self.status[u]
            eventp = np.random.random_sample()
            neighbors = self.graph.neighbors(u)
            if self.graph.directed:
                neighbors = self.graph.predecessors(u)

            if u_status == 0:  # Susceptible
                infected_neighbors = [v for v in neighbors if self.status[v] == 1]
                triggered = 1 if len(infected_neighbors) > 0 else 0

                if self.params['model']['tp_rate'] == 1:
                    if eventp < 1 - (1 - self.params['model']['beta']) ** len(infected_neighbors):
                        actual_status[u] = 2  # Exposed
                        self.progress[u] = 0
                else:
                    if eventp < self.params['model']['beta'] * triggered:
                        actual_status[u] = 2  # Exposed
                        self.progress[u] = 0

            elif u_status == 2:

                # apply prob. of infection, after (t - t_i)
                if eventp < 1 - np.exp(- (self.actual_iteration - self.progress[u]) * self.params['model']['alpha']):
                    actual_status[u] = 1  # Infected
                    self.progress_I[u] = self.actual_iteration
                    del self.progress[u]

            elif u_status == 1:

                if eventp < 1 - np.exp(- (self.actual_iteration - self.progress_I[u]) * self.params['model']['lambda']):
                    actual_status[u] = 0  # Removed
                    del self.progress_I[u]

        delta, node_count, status_delta = self.status_delta(actual_status)
        self.status = actual_status
        self.actual_iteration += 1

        if node_status:
            return {"iteration": self.actual_iteration - 1, "status": delta.copy(),
                    "node_count": node_count.copy(), "status_delta": status_delta.copy()}
        else:
            return {"iteration": self.actual_iteration - 1, "status": {},
                    "node_count": node_count.copy(), "status_delta": status_delta.copy()}



from ..DiffusionModel import DiffusionModel
import future.utils

__author__ = "Giulio Rossetti"
__license__ = "BSD-2-Clause"
__email__ = "giulio.rossetti@gmail.com"


class ThresholdModel(DiffusionModel):
    """
        Node Parameters to be specified via ModelConfig

       :param threshold: The node threshold. If not specified otherwise a value of 0.1 is assumed for all nodes.
    """

    def __init__(self, graph, seed=None):
        """
             Model Constructor

             :param graph: A networkx graph object
         """
        super(self.__class__, self).__init__(graph, seed)
        self.available_statuses = {
            "Susceptible": 0,
            "Infected": 1
        }

        self.parameters = {
            "model": {},
            "nodes": {
                "threshold": {
                    "descr": "Node threshold",
                    "range": [0, 1],
                    "optional": True,
                    "default": 0.1
                }
            },
            "edges": {},
        }

        self.name = "Threshold"

    def iteration(self, node_status=True):
        """
        Execute a single model iteration

        :return: Iteration_id, Incremental node status (dictionary node->status)
        """
        self.clean_initial_status(self.available_statuses.values())

        actual_status = {node: nstatus for node, nstatus in future.utils.iteritems(self.status)}

        if self.actual_iteration == 0:
            self.actual_iteration += 1
            delta, node_count, status_delta = self.status_delta(actual_status)
            if node_status:
                return {"iteration": 0, "status": actual_status.copy(),
                        "node_count": node_count.copy(), "status_delta": status_delta.copy()}
            else:
                return {"iteration": 0, "status": {},
                        "node_count": node_count.copy(), "status_delta": status_delta.copy()}

        for u in self.graph.nodes:
            if actual_status[u] == 1:
                continue

            neighbors = list(self.graph.neighbors(u))
            if self.graph.directed:
                neighbors = list(self.graph.predecessors(u))

            infected = 0
            for v in neighbors:
                infected += self.status[v]

            if len(neighbors) > 0:
                infected_ratio = float(infected)/len(neighbors)
                if infected_ratio >= self.params['nodes']['threshold'][u]:
                    actual_status[u] = 1

        delta, node_count, status_delta = self.status_delta(actual_status)
        self.status = actual_status
        self.actual_iteration += 1

        if node_status:
            return {"iteration": self.actual_iteration - 1, "status": delta.copy(),
                    "node_count": node_count.copy(), "status_delta": status_delta.copy()}
        else:
            return {"iteration": self.actual_iteration - 1, "status": {},
                    "node_count": node_count.copy(), "status_delta": status_delta.copy()}



import future.utils

__author__ = 'Giulio Rossetti'
__license__ = "BSD-2-Clause"
__email__ = "giulio.rossetti@gmail.com"


class Configuration(object):
    """
    Configuration Object

    """

    def __init__(self):
        self.config = {
            'nodes': {},
            'edges': {},
            'model': {},
            'status': {}
        }

    def get_nodes_configuration(self):
        """
        Nodes configurations

        :return: dictionary that link each node to its attributes
        """
        return self.config['nodes']

    def get_edges_configuration(self):
        """
        Edges configurations

        :return: dictionary that link each edge to its attributes
        """
        return self.config['edges']

    def get_model_parameters(self):
        """
        Model parameters

        :return: dictionary describes the specified model parameters
        """
        return self.config['model']

    def get_model_configuration(self):
        """
        Initial configuration

        :return: initial nodes status (if specified)
        """
        return self.config['status']

    def add_model_parameter(self, param_name, param_value):
        """
        Set a Model Parameter

        :param param_name: parameter identifier (as specified by the chosen model)
        :param param_value: parameter value
        """
        self.config['model'][param_name] = param_value

    def add_model_initial_configuration(self, status_name, nodes):
        """
        Set initial status for a set of nodes

        :param status_name: status to be set (as specified by the chosen model)
        :param nodes: list of affected nodes
        """
        self.config['status'][status_name] = nodes

    def add_node_configuration(self, param_name, node_id, param_value):
        """
        Set a parameter for a given node

        :param param_name: parameter identifier (as specified by the chosen model)
        :param node_id: node identifier
        :param param_value: parameter value
        """
        if param_name not in self.config['nodes']:
            self.config['nodes'][param_name] = {node_id: param_value}
        else:
            self.config['nodes'][param_name][node_id] = param_value

    def add_node_set_configuration(self, param_name, node_to_value):
        """
        Set Nodes parameter

        :param param_name: parameter identifier (as specified by the chosen model)
        :param node_to_value: dictionary mapping each node a parameter value
        """
        for nid, val in future.utils.iteritems(node_to_value):
            self.add_node_configuration(param_name, nid, val)

    def add_edge_configuration(self, param_name, edge, param_value):
        """
        Set a parameter for a given edge

        :param param_name: parameter identifier (as specified by the chosen model)
        :param edge: edge identifier
        :param param_value: parameter value
        """
        if param_name not in self.config['edges']:
            self.config['edges'][param_name] = {edge: param_value}
        else:
            self.config['edges'][param_name][edge] = param_value

    def add_edge_set_configuration(self, param_name, edge_to_value):
        """
        Set Edges parameter

        :param param_name: parameter identifier (as specified by the chosen model)
        :param edge_to_value: dictionary mapping each edge a parameter value
        """
        for edge, val in future.utils.iteritems(edge_to_value):
            self.add_edge_configuration(param_name, edge, val)



import numpy as np
import multiprocessing
from contextlib import closing
import copy
import past

__author__ = 'Giulio Rossetti'
__license__ = "BSD-2-Clause"
__email__ = "giulio.rossetti@gmail.com"


class InitializationException(Exception):
    """Initialization Exception"""


def multi_runs(model, execution_number=1, iteration_number=50, infection_sets=None,
               nprocesses=multiprocessing.cpu_count()):
    """
    Multiple executions of a given model varying the initial set of infected nodes

    :param model: a configured diffusion model
    :param execution_number: number of instantiations
    :param iteration_number: number of iterations per execution
    :param infection_sets: predefined set of infected nodes sets
    :param nprocesses: number of processes. Default values cpu number.
    :return: resulting trends for all the executions
    """

    if nprocesses > multiprocessing.cpu_count():
        nprocesses = multiprocessing.cpu_count()

    executions = []
    seeds = np.around(np.random.rand(execution_number)*2**32).astype(int)

    if infection_sets is not None:
        if len(infection_sets) != execution_number:
            raise InitializationException(
                {"message": "Number of infection sets provided does not match the number of executions required"})

        for x in past.builtins.xrange(0, execution_number, nprocesses):

            with closing(multiprocessing.Pool(processes=nprocesses, maxtasksperchild=10)) as pool:
                tasks = [(seeds[i], copy.deepcopy(model).reset(infection_sets[i])) for i in
                         past.builtins.xrange(x, min(x + nprocesses, execution_number))]
                results = [pool.apply_async(__execute, (*t, iteration_number)) for t in tasks]

            for result in results:
                executions.append(result.get())
    else:
        for x in past.builtins.xrange(0, execution_number, nprocesses):
            with closing(multiprocessing.Pool(processes=nprocesses, maxtasksperchild=10)) as pool:

                tasks = [(seeds[i], copy.deepcopy(model).reset()) for i in
                         past.builtins.xrange(x, min(x + nprocesses, execution_number))]
                results = [pool.apply_async(__execute, (*t, iteration_number)) for t in tasks]

            for result in results:
                executions.append(result.get())

    return executions


def __execute(seed, model, iteration_number):
    """
    Execute a simulation model

    :param model: a configured diffusion model
    :param iteration_number: number of iterations
    :return: computed trends
    """
    np.random.seed(seed)
    iterations = model.iteration_bunch(iteration_number, False)
    trends = model.build_trends(iterations)[0]
    del iterations
    del model
    return trends






__author__ = 'Giulio Rossetti'
__license__ = "BSD-2-Clause"
__email__ = "giulio.rossetti@gmail.com"

import ndlib.parser.ExperimentParser as ep
import argparse


def translate():

    parser = argparse.ArgumentParser()

    parser.add_argument('query_file', type=str, help='simulation query file')
    parser.add_argument('python_file', type=str, help='Python script file')

    args = parser.parse_args()
    query = ""
    with open(args.query_file) as f:
        query = f.read()

    NDQL_parser = ep.ExperimentParser()
    NDQL_parser.set_query(query)
    NDQL_parser.parse()

    with open(args.python_file, "w") as o:
        o.write(NDQL_parser.script)




from ndlib.viz.mpl.ComparisonViz import ComparisonPlot
import numpy as np

__author__ = 'Giulio Rossetti'
__license__ = "BSD-2-Clause"
__email__ = "giulio.rossetti@gmail.com"


class DiffusionPrevalenceComparison(ComparisonPlot):

    def __init__(self, models, trends, statuses=("Infected")):
        """
        :param models: A list of model object
        :param trends: A list of computed simulation trends
        :param statuses: The model statuses for which make the plot. Default ["Infected"].
        """
        super(self.__class__, self).__init__(models, trends, statuses)
        self.ylabel = "#Nodes"
        self.title = "DPrevalence"
        self.normalized = False

    def iteration_series(self, percentile):

        mseries = {mn: {} for mn in self.mnames}

        i = 0
        for trend in self.trends:
            presences = {k: [] for k in self.classes}
            for t in trend:
                for k in t['trends']['status_delta'].keys():
                    if self.srev[self.mnames[i]][k] in presences:
                        presences[self.srev[self.mnames[i]][k]].append(np.array(t['trends']['status_delta'][k]))

            for st in presences:
                tp = np.percentile(np.array(presences[st]), percentile, axis=0)
                bp = np.percentile(np.array(presences[st]), 100-percentile, axis=0)
                av = np.average(np.array(presences[st]), axis=0)
                if self.mnames[i] not in mseries:
                    mseries[self.mnames[i]] = {st: (tp, av, bp)}
                else:
                    mseries[self.mnames[i]][st] = (tp, av, bp)
            i += 1

        return mseries



from ndlib.models.compartments.Compartment import Compartiment
import networkx as nx
import numpy as np

__author__ = 'Giulio Rossetti'
__license__ = "BSD-2-Clause"
__email__ = "giulio.rossetti@gmail.com"


class EdgeCategoricalAttribute(Compartiment):

    def __init__(self, attribute, value, triggering_status=None, probability=1, **kwargs):
        super(self.__class__, self).__init__(kwargs)
        self.attribute = attribute

        if not isinstance(value, str):
            raise ValueError("Categorical (string) value expected")

        self.attribute_value = value
        self.trigger = triggering_status
        self.probability = probability

    def execute(self, node, graph, status, status_map, *args, **kwargs):
        neighbors = list(graph.neighbors(node))
        if isinstance(graph, nx.DiGraph):
            neighbors = list(graph.predecessors(node))

        edge_attr = graph.get_edge_attributes(self.attribute)

        if self.trigger is not None:
            triggered = [v for v in neighbors if status[v] == status_map[self.trigger] and
                         edge_attr[(min([node, v]), max([node, v]))] == self.attribute_value]
        else:
            triggered = [v for v in neighbors if edge_attr[(min([node, v]), max([node, v]))] == self.attribute_value]

        for _ in triggered:
            p = np.random.random_sample()
            test = p <= self.probability
            if test:
                return self.compose(node, graph, status, status_map, kwargs)

        return False



from ndlib.models.DynamicDiffusionModel import DynamicDiffusionModel
import future.utils

__author__ = 'Giulio Rossetti'
__license__ = "BSD-2-Clause"
__email__ = "giulio.rossetti@gmail.com"


class DynamicCompositeModel(DynamicDiffusionModel):

    def __init__(self, graph):
        """
             Model Constructor
             :param graph: A networkx graph object
         """
        super(self.__class__, self).__init__(graph)
        self.available_statuses = {}
        self.compartment = {}
        self.compartment_progressive = 0
        self.status_progressive = 0

    def add_status(self, status_name):
        self.available_statuses[status_name] = self.status_progressive
        self.status_progressive += 1

    def add_rule(self, status_from, status_to, rule):
        self.compartment[self.compartment_progressive] = (status_from, status_to, rule)
        self.compartment_progressive += 1

    def iteration(self, node_status=True):
        """
        Execute a single model iteration

        :return: Iteration_id, Incremental node status (dictionary node->status)
        """
        self.clean_initial_status(self.available_statuses.values())
        actual_status = {node: nstatus for node, nstatus in future.utils.iteritems(self.status)}

        if self.actual_iteration == 0:
            self.actual_iteration += 1
            delta, node_count, status_delta = self.status_delta(actual_status)
            if node_status:
                return {"iteration": 0, "status": actual_status.copy(),
                        "node_count": node_count.copy(), "status_delta": status_delta.copy()}
            else:
                return {"iteration": 0, "status": {},
                        "node_count": node_count.copy(), "status_delta": status_delta.copy()}

        for u in self.graph.nodes():
            u_status = self.status[u]
            for i in range(0, self.compartment_progressive):

                if u_status == self.available_statuses[self.compartment[i][0]]:
                    rule = self.compartment[i][2]
                    test = rule.execute(node=u, graph=self.graph, status=self.status,
                                        status_map=self.available_statuses, params=self.params)
                    if test:
                        actual_status[u] = self.available_statuses[self.compartment[i][1]]
                        break

        delta, node_count, status_delta = self.status_delta(actual_status)
        self.status = actual_status
        self.actual_iteration += 1

        if node_status:
            return {"iteration": self.actual_iteration - 1, "status": delta.copy(),
                    "node_count": node_count.copy(), "status_delta": status_delta.copy()}
        else:
            return {"iteration": self.actual_iteration - 1, "status": {},
                    "node_count": node_count.copy(), "status_delta": status_delta.copy()}



from ..DiffusionModel import DiffusionModel
import numpy as np

__author__ = "Alina Sirbu"
__email__ = "alina.sirbu@unipi.it"


class VoterModel(DiffusionModel):
    """

    """

    def __init__(self, graph, seed=None):
        """
             Model Constructor

             :param graph: A networkx graph object
         """
        super(self.__class__, self).__init__(graph, seed)
        self.available_statuses = {
            "Susceptible": 0,
            "Infected": 1
        }

        self.name = "Voter"

    def iteration(self, node_status=True):
        """
        Execute a single model iteration

        :return: Iteration_id, Incremental node status (dictionary node->status)
        """
        # One iteration changes the opinion of one voter using the following procedure:
        # - select randomly one voter (listener)
        # - selecting randomly one of its peers (speaker)
        # - the first voter takes the opinion of the peer (listener takes the opinion of speaker)

        self.clean_initial_status(self.available_statuses.values())

        if self.actual_iteration == 0:
            self.actual_iteration += 1

            delta, node_count, status_delta = self.status_delta(self.status)
            if node_status:
                return {"iteration": 0, "status": self.status.copy(),
                        "node_count": node_count.copy(), "status_delta": status_delta.copy()}
            else:
                return {"iteration": 0, "status": {},
                        "node_count": node_count.copy(), "status_delta": status_delta.copy()}

        # select a random node
        listener = list(self.graph.nodes)[np.random.randint(0, self.graph.number_of_nodes())]

        # select a random neighbour
        neighbours = list(self.graph.neighbors(listener))
        if self.graph.directed:
            # difficult to have a digraph but assumed if a->b then b can be influenced by a
            # but not the other way around
            neighbours = list(self.graph.predecessors(listener))

        speaker = neighbours[np.random.randint(0, len(neighbours))]

        # update status of listener
        delta = {listener: self.status[speaker]}
        self.status[listener] = self.status[speaker]

        # fix
        node_count = {st: len([n for n in self.status if self.status[n] == st])
                      for st in self.available_statuses.values()}
        status_delta = {st: 0 for st in self.available_statuses.values()}
        status_delta[self.status[speaker]] += 1
        for x in self.available_statuses.values():
            if x != self.status[speaker]:
                status_delta[x] -= 1

        self.actual_iteration += 1

        if node_status:
            return {"iteration": self.actual_iteration - 1, "status": delta.copy(),
                    "node_count": node_count.copy(), "status_delta": status_delta.copy()}
        else:
            return {"iteration": self.actual_iteration - 1, "status": {},
                    "node_count": node_count.copy(), "status_delta": status_delta.copy()}




"""
The :mod:`ndlib.models` module contains...
"""

from . import dynamic
from .dynamic import DynSIModel
from .dynamic import DynSIRModel
from .dynamic import DynSISModel
from .dynamic import DynKerteszThresholdModel
from .dynamic import DynProfileModel
from .dynamic import DynProfileThresholdModel



import abc
import warnings
import numpy as np
import past.builtins
import future.utils
import six
from netdispatch import AGraph
import tqdm

__author__ = "Giulio Rossetti"
__license__ = "BSD-2-Clause"
__email__ = "giulio.rossetti@gmail.com"


class ConfigurationException(Exception):
    """Configuration Exception"""


@six.add_metaclass(abc.ABCMeta)
class DiffusionModel(object):
    """
        Partial Abstract Class that defines Diffusion Models
    """
    # __metaclass__ = abc.ABCMeta

    def __init__(self, graph, seed=None):
        """
            Model Constructor

            :param graph: A networkx graph object
        """

        np.random.seed(seed)

        self.discrete_state = True

        self.params = {
            'nodes': {},
            'edges': {},
            'model': {},
            'status': {}
        }

        self.available_statuses = {
            "Susceptible": 0,
            "Infected": 1,
            "Recovered": 2,
            "Blocked": -1
        }

        self.name = ""

        self.parameters = {
            "model": {},
            "nodes": {},
            "edges": {}
        }

        self.actual_iteration = 0
        self.graph = AGraph(graph)
        self.status = {n: 0 for n in self.graph.nodes}
        self.initial_status = {}

    def __validate_configuration(self, configuration):
        """
        Validate the consistency of a Configuration object for the specific model

        :param configuration: a Configuration object instance
        """
        if "Infected" not in self.available_statuses:
            raise ConfigurationException("'Infected' status not defined.")

        # Checking mandatory parameters
        omp = set([k for k in self.parameters['model'].keys() if not self.parameters['model'][k]['optional']])
        onp = set([k for k in self.parameters['nodes'].keys() if not self.parameters['nodes'][k]['optional']])
        oep = set([k for k in self.parameters['edges'].keys() if not self.parameters['edges'][k]['optional']])

        mdp = set(configuration.get_model_parameters().keys())
        ndp = set(configuration.get_nodes_configuration().keys())
        edp = set(configuration.get_edges_configuration().keys())

        if len(omp) > 0:
            if len(omp & mdp) != len(omp):
                raise ConfigurationException({"message": "Missing mandatory model parameter(s)", "parameters": omp-mdp})

        if len(onp) > 0:
            if len(onp & ndp) != len(onp):
                raise ConfigurationException({"message": "Missing mandatory node parameter(s)", "parameters": onp-ndp})

        if len(oep) > 0:
            if len(oep & edp) != len(oep):
                raise ConfigurationException({"message": "Missing mandatory edge parameter(s)", "parameters": oep-edp})

        # Checking optional parameters
        omp = set([k for k in self.parameters['model'].keys() if self.parameters['model'][k]['optional']])
        onp = set([k for k in self.parameters['nodes'].keys() if self.parameters['nodes'][k]['optional']])
        oep = set([k for k in self.parameters['edges'].keys() if self.parameters['edges'][k]['optional']])

        if len(omp) > 0:
            for param in omp:
                if param not in mdp:
                    configuration.add_model_parameter(param, self.parameters['model'][param]['default'])

        if len(onp) > 0:
            for param in onp:
                if param not in ndp:
                    for nid in self.graph.nodes:
                        configuration.add_node_configuration(param, nid, self.parameters['nodes'][param]['default'])

        if len(oep) > 0:
            for param in oep:
                if param not in edp:
                    for eid in self.graph.edges:
                        configuration.add_edge_configuration(param, eid, self.parameters['edges'][param]['default'])

        # Checking initial simulation status
        sts = set(configuration.get_model_configuration().keys())
        if self.discrete_state and "Infected" not in sts and "fraction_infected" not in mdp \
                and "percentage_infected" not in mdp:
            warnings.warn('Initial infection missing: a random sample of 5% of graph nodes will be set as infected')
            self.params['model']["fraction_infected"] = 0.05

    def set_initial_status(self, configuration):
        """
        Set the initial model configuration

        :param configuration: a ```ndlib.models.ModelConfig.Configuration``` object
        """

        self.__validate_configuration(configuration)

        nodes_cfg = configuration.get_nodes_configuration()
        # Set additional node information

        for param, node_to_value in future.utils.iteritems(nodes_cfg):
            if len(node_to_value) < len(self.graph.nodes):
                raise ConfigurationException({"message": "Not all nodes have a configuration specified"})

            self.params['nodes'][param] = node_to_value

        edges_cfg = configuration.get_edges_configuration()
        # Set additional edges information
        for param, edge_to_values in future.utils.iteritems(edges_cfg):
            if len(edge_to_values) == len(self.graph.edges):
                self.params['edges'][param] = {}
                for e in edge_to_values:
                    self.params['edges'][param][e] = edge_to_values[e]

        # Set initial status
        model_status = configuration.get_model_configuration()

        for param, nodes in future.utils.iteritems(model_status):
            self.params['status'][param] = nodes
            for node in nodes:
                self.status[node] = self.available_statuses[param]

        # Set model additional information
        model_params = configuration.get_model_parameters()
        for param, val in future.utils.iteritems(model_params):
            self.params['model'][param] = val

        # Handle initial infection
        if 'Infected' not in self.params['status']:
            if 'percentage_infected' in self.params['model']:
                self.params['model']['fraction_infected'] = self.params['model']['percentage_infected']
            if 'fraction_infected' in self.params['model']:
                number_of_initial_infected = self.graph.number_of_nodes() * float(self.params['model']['fraction_infected'])
                if number_of_initial_infected < 1:
                    warnings.warn(
                        "The fraction_infected value is too low given the number of nodes of the selected graph: a "
                        "single node will be set as infected")
                    number_of_initial_infected = 1

                available_nodes = [n for n in self.status if self.status[n] == 0]
                sampled_nodes = np.random.choice(available_nodes, int(number_of_initial_infected), replace=False)
                for k in sampled_nodes:
                    self.status[k] = self.available_statuses['Infected']

        self.initial_status = self.status

    def clean_initial_status(self, valid_status=None):
        """
        Check the consistency of initial status
        :param valid_status: valid node configurations
        """
        for n, s in future.utils.iteritems(self.status):
            if s not in valid_status:
                self.status[n] = 0

    def iteration_bunch(self, bunch_size, node_status=True):
        """
        Execute a bunch of model iterations

        :param bunch_size: the number of iterations to execute
        :param node_status: if the incremental node status has to be returned.

        :return: a list containing for each iteration a dictionary {"iteration": iteration_id, "status": dictionary_node_to_status}
        """
        system_status = []
        for it in tqdm.tqdm(past.builtins.xrange(0, bunch_size)):
            its = self.iteration(node_status)
            system_status.append(its)
        return system_status

    def get_info(self):
        """
        Describes the current model parameters (nodes, edges, status)

        :return: a dictionary containing for each parameter class the values specified during model configuration
        """
        info = {k: v for k, v in future.utils.iteritems(self.params) if k not in ['nodes', 'edges', 'status']}
        if 'infected_nodes' in self.params['status']:
            info['selected_initial_infected'] = True
        return info['model']

    def reset(self, infected_nodes=None):
        """
        Reset the simulation setting the actual status to the initial configuration.
        """
        self.actual_iteration = 0

        if infected_nodes is not None:
            for n in self.status:
                self.status[n] = 0
            for n in infected_nodes:
                self.status[n] = self.available_statuses['Infected']
            self.initial_status = self.status

        else:
            if 'percentage_infected' in self.params['model']:
                self.params['model']['fraction_infected'] = self.params['model']['percentage_infected']
            if 'fraction_infected' in self.params['model']:
                for n in self.status:
                    self.status[n] = 0
                number_of_initial_infected = self.graph.number_of_nodes() * float(self.params['model']['fraction_infected'])
                available_nodes = [n for n in self.status if self.status[n] == 0]
                sampled_nodes = np.random.choice(available_nodes, int(number_of_initial_infected), replace=False)

                for k in sampled_nodes:
                    self.status[k] = self.available_statuses['Infected']

                self.initial_status = self.status
            else:
                self.status = self.initial_status

        return self

    def get_model_parameters(self):
        return self.parameters

    def get_name(self):
        return self.name

    def get_status_map(self):
        """
        Specify the statuses allowed by the model and their numeric code

        :return: a dictionary (status->code)
        """
        return self.available_statuses

    @abc.abstractmethod
    def iteration(self, node_status=True):
        """
        Execute a single model iteration

        :param node_status: if the incremental node status has to be returned.

        :return: Iteration_id,
                 (optional) Incremental node status (dictionary node->status),
                 Status count (dictionary status->node count),
                 Status delta (dictionary status->node delta)
        """
        pass

    @staticmethod
    def check_status_similarity(actual, previous):
        """
        Evaluate similarity among statuses

        :param actual: actual status
        :param previous: previous status
        :return: True if the two statuses are the same, False otherwise
        """
        for n, v in future.utils.iteritems(actual):
            if n not in previous:
                return False
            if previous[n] != actual[n]:
                return False
        return True

    def status_delta(self, actual_status):
        """
        Compute the point-to-point variations for each status w.r.t. the previous system configuration

        :param actual_status: the actual simulation status
        :return: node that have changed their statuses (dictionary status->nodes),
                 count of actual nodes per status (dictionary status->node count),
                 delta of nodes per status w.r.t the previous configuration (dictionary status->delta)
        """
        actual_status_count = {}
        old_status_count = {}
        delta = {}
        for n, v in future.utils.iteritems(self.status):
            if v != actual_status[n]:
                delta[n] = actual_status[n]

        for st in list(self.available_statuses.values()):
            actual_status_count[st] = len([x for x in actual_status if actual_status[x] == st])
            old_status_count[st] = len([x for x in self.status if self.status[x] == st])

        status_delta = {st: actual_status_count[st] - old_status_count[st] for st in actual_status_count}

        return delta, actual_status_count, status_delta

    def status_delta_continuous(self, actual_status):
        """
        Compute the point-to-point variations for each status w.r.t. the previous system configuration

        Should be used for continuous statuses instead of discrete values

        :param actual_status: the actual simulation status
        :return: nodes that have changed their statuses (dictionary status->nodes),
                    count of actual nodes per status (dictionary status->node count),
                    delta of nodes per status w.r.t the previous configuration (dictionary status->delta)
        """
        delta = {}
        status_delta = {}
        for n, v in future.utils.iteritems(self.status):
            delta[n] = {}
            status_delta[n] = {}
            for var, val in v.items():
                if val != actual_status[n][var]:
                    delta[n][var] = actual_status[n][var]
                    status_delta[n][var] = actual_status[n][var] - val
            if len(delta[n].values()) == 0:
                del delta[n]
            if len(status_delta[n].values()) == 0:
                del status_delta[n]

        return delta, status_delta


    def build_trends(self, iterations):
        """
        Build node status and node delta trends from model iteration bunch

        :param iterations: a set of iterations
        :return: a trend description
        """
        status_delta = {status: [] for status in self.available_statuses.values()}
        node_count = {status: [] for status in self.available_statuses.values()}

        for it in iterations:
            for st in self.available_statuses.values():
                try:
                    status_delta[st].append(it['status_delta'][st])
                    node_count[st].append(it['node_count'][st])
                except:
                    status_delta[st].append(it['status_delta'][str(st)])
                    node_count[st].append(it['node_count'][str(st)])

        return [{"trends": {"node_count": node_count, "status_delta": status_delta}}]
